from morpfw.crud import Schema, Collection, Model
from .schema import APIKeySchema
import jsl
from ..app import App
from uuid import uuid4
import rulez


class APIKeyModel(Model):
    schema = APIKeySchema


class APIKeyCollection(Collection):
    schema = APIKeySchema

    def create(self, data):
        if not data.get('userid', None):
            data['userid'] = self.request.identity.userid
        data['api_identity'] = uuid4().hex
        data['api_secret'] = uuid4().hex
        return super().create(data)

    def search(self, query=None, *args, **kwargs):
        if kwargs.get('secure', True):
            if query:
                rulez.and_(
                    rulez.field['userid'] == self.request.identity.userid,
                    query)
            else:
                query = rulez.field['userid'] == self.request.identity.userid
        return super(APIKeyCollection, self).search(query, *args, **kwargs)



import dectate
import reg
import morepath
from .blobstorage.base import BlobStorage
from . import signals


class StorageAction(dectate.Action):

    app_class_arg = True

    def __init__(self, model):
        self.model = model

    def identifier(self, app_class):
        return str((self.model))

    def perform(self, obj, app_class):
        app_class._get_storage.register(
            reg.methodify(obj), model=self.model,
            request=morepath.Request, blobstorage=BlobStorage)


class BlobStorageAction(dectate.Action):

    app_class_arg = True

    def __init__(self, model):
        self.model = model

    def identifier(self, app_class):
        return str((self.model))

    def perform(self, obj, app_class):
        app_class._get_blobstorage.register(
            reg.methodify(obj), model=self.model, request=morepath.Request)


class DataProviderAction(dectate.Action):

    app_class_arg = True

    def __init__(self, schema, obj, storage):
        self.schema = schema
        self.obj = obj
        self.storage = storage

    def identifier(self, app_class):
        return str((app_class, self.schema, self.obj, self.storage))

    def perform(self, obj, app_class):
        app_class.get_dataprovider.register(reg.methodify(obj),
                                            schema=self.schema,
                                            obj=self.obj,
                                            storage=self.storage)


class JSONProviderAction(dectate.Action):

    app_class_arg = True

    def __init__(self, obj):
        self.obj = obj

    def identifier(self, app_class):
        return str((app_class, self.obj))

    def perform(self, obj, app_class):
        app_class.get_jsonprovider.register(
            reg.methodify(obj), obj=self.obj)


class IdentifierFieldsAction(dectate.Action):

    app_class_arg = True

    def __init__(self, schema):
        self.schema = schema

    def identifier(self, app_class):
        return str((app_class, self.schema))

    def perform(self, obj, app_class):
        app_class.get_identifierfields.register(
            reg.methodify(obj), schema=self.schema)


class UUIDFieldAction(dectate.Action):

    app_class_arg = True

    def __init__(self, schema):
        self.schema = schema

    def identifier(self, app_class):
        return str((app_class, self.schema))

    def perform(self, obj, app_class):
        app_class.get_uuidfield.register(
            reg.methodify(obj), schema=self.schema)


class DefaultIdentifierAction(dectate.Action):

    app_class_arg = True

    def __init__(self, schema):
        self.schema = schema

    def identifier(self, app_class):
        return str((app_class, self.schema))

    def perform(self, obj, app_class):
        app_class.get_default_identifier.register(
            reg.methodify(obj), schema=self.schema)


class FormValidatorAction(dectate.Action):

    app_class_arg = True

    def __init__(self, schema):
        self.schema = schema

    def identifier(self, app_class):
        return str((app_class, self.schema))

    def perform(self, obj, app_class):
        app_class.get_formvalidators.register(
            reg.methodify(obj), schema=self.schema)


class RulesAdapterAction(dectate.Action):

    app_class_arg = True

    def __init__(self, model):
        self.model = model

    def identifier(self, app_class):
        return str((app_class, self.model))

    def perform(self, obj, app_class):
        app_class.get_rulesadapter.register(
            reg.methodify(obj), model=self.model)


class StateMachineAction(dectate.Action):

    app_class_arg = True

    def __init__(self, model):
        self.model = model

    def identifier(self, app_class):
        return str((app_class, self.model))

    def perform(self, obj, app_class):
        app_class.get_statemachine.register(
            reg.methodify(obj), model=self.model)


class XattrProviderAction(dectate.Action):

    app_class_arg = True

    def __init__(self, model):
        self.model = model

    def identifier(self, app_class):
        return str((app_class, self.model))

    def perform(self, obj, app_class):
        app_class.get_xattrprovider.register(
            reg.methodify(obj), model=self.model)



from .crud_common import App as BaseApp
from morpfw.crud.model import Collection, Model
from morpfw.crud.schema import Schema
from morpfw.crud.storage.elasticsearchstorage import ElasticSearchStorage
from morpfw.crud.blobstorage.fsblobstorage import FSBlobStorage
from .crud_common import get_client, run_jslcrud_test, PageCollection, PageModel
from .crud_common import NamedObjectCollection, NamedObjectModel
from .crud_common import BlobObjectCollection, BlobObjectModel
from .crud_common import ObjectXattrProvider, ObjectXattrSchema
from .crud_common import FSBLOB_DIR
from .crud_common import ObjectSchema as BaseObjectSchema
import pprint
from more.transaction import TransactionApp
from morepath.reify import reify
from morepath.request import Request
from more.basicauth import BasicAuthIdentityPolicy
from sqlalchemy.orm import sessionmaker
from zope.sqlalchemy import register as register_session
import morpfw.crud.signals as signals
from elasticsearch import Elasticsearch
from dataclasses import dataclass, field
import typing

Session = sessionmaker()
register_session(Session)


class ESClientRequest(Request):

    @reify
    def es_client(self):
        return Elasticsearch(hosts='127.0.0.1:9085')


class App(BaseApp, TransactionApp):

    request_class = ESClientRequest


class PageStorage(ElasticSearchStorage):
    index_name = 'test-index'
    doc_type = 'page'
    refresh = 'wait_for'
    model = PageModel


@App.path(model=PageCollection, path='pages')
def collection_factory(request):
    storage = PageStorage(request)
    return PageCollection(request, storage)


@App.path(model=PageModel, path='pages/{identifier}')
def model_factory(request, identifier):
    storage = PageStorage(request)
    return storage.get(identifier)


@dataclass
class ObjectSchema(BaseObjectSchema):

    id: typing.Optional[str] = None


@App.identifierfields(schema=ObjectSchema)
def object_identifierfields(schema):
    return ['id']


@App.default_identifier(schema=ObjectSchema)
def object_default_identifier(schema, obj, request):
    return None


class ObjectModel(Model):
    schema = ObjectSchema


@App.xattrprovider(model=ObjectModel)
def get_xattr_provider(context):
    return ObjectXattrProvider(context)


class ObjectCollection(Collection):
    schema = ObjectSchema


@App.subscribe(signal=signals.OBJECT_CREATED, model=ObjectModel)
def object_created(app, request, obj, signal):
    obj.data['created_flag'] = True


@App.subscribe(signal=signals.OBJECT_UPDATED, model=ObjectModel)
def object_updated(app, request, obj, signal):
    obj.data['updated_flag'] = True


class ObjectStorage(ElasticSearchStorage):
    index_name = 'test-index'
    doc_type = 'object'
    refresh = 'wait_for'
    auto_id = True
    model = ObjectModel


@App.json(model=ObjectCollection, name='get_uuid')
def get_object_by_uuid(context, request):
    uuid = request.GET.get('uuid')
    return context.get_by_uuid(uuid).json()


@App.path(model=ObjectCollection, path='objects')
def object_collection_factory(request):
    storage = ObjectStorage(request)
    return ObjectCollection(request, storage)


@App.path(model=ObjectModel, path='objects/{identifier}')
def object_model_factory(request, identifier):
    storage = ObjectStorage(request)
    return storage.get(identifier)


class NamedObjectStorage(ElasticSearchStorage):
    index_name = 'test-index'
    doc_type = 'namedobject'
    refresh = 'wait_for'
    model = NamedObjectModel


@App.path(model=NamedObjectCollection, path='named_objects')
def namedobject_collection_factory(request):
    storage = NamedObjectStorage(request)
    return NamedObjectCollection(request, storage)


@App.path(model=NamedObjectModel, path='named_objects/{identifier}')
def namedobject_model_factory(request, identifier):
    storage = NamedObjectStorage(request)
    return storage.get(identifier)


class BlobObjectStorage(ElasticSearchStorage):
    index_name = 'test-index'
    doc_type = 'blobobject'
    refresh = 'wait_for'
    model = BlobObjectModel


@App.storage(model=BlobObjectModel)
def get_blobobject_storage(model, request, blobstorage):
    return BlobObjectStorage(request, blobstorage=blobstorage)


@App.blobstorage(model=BlobObjectModel)
def get_blobobject_blobstorage(model, request):
    return FSBlobStorage(request, FSBLOB_DIR)


@App.path(model=BlobObjectCollection, path='blob_objects')
def blobobject_collection_factory(request):
    storage = request.app.get_storage(BlobObjectModel, request)
    return BlobObjectCollection(request, storage)


@App.path(model=BlobObjectModel, path='blob_objects/{identifier}')
def blobobject_model_factory(request, identifier):
    storage = request.app.get_storage(BlobObjectModel, request)
    return storage.get(identifier)


def test_elasticsearchstorage(es_client):
    es_client.indices.create('test-index', body={
        'settings': {
            'number_of_shards': 1,
            'number_of_replicas': 0
        }
    })
    es_client.transport.perform_request('PUT', '/test-index/_mapping/page',
                                        body={
                                            'properties': {
                                                'title': {
                                                    'type': 'text',
                                                    'fielddata': True
                                                }
                                            }
                                        })
    client = get_client(App)
    run_jslcrud_test(client)



import morepath
import dectate
import reg
from .crud.app import App as CRUDApp
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.pool import NullPool, QueuePool
from more.transaction import TransactionApp
from morepath.reify import reify
from morepath.request import Request as BaseRequest
import webob
from more import cors
from morepath.request import Response
from . import directive
from . import exc
from celery import Celery, Task
from celery import shared_task
from celery.schedules import crontab
import time
import re
import sqlalchemy
from .sql import Base
import transaction
import os
from zope.sqlalchemy import register as register_session
import transaction
from zope.sqlalchemy import ZopeTransactionExtension
from .exc import ConfigurationError
import warnings
import sqlalchemy.orm
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
#from typing import Union
from celery.local import Proxy
from celery.result import AsyncResult
from sqlalchemy.engine.base import Engine
from billiard.einfo import ExceptionInfo
from .signal.app import SignalApp
import reg

Session = sessionmaker(extension=ZopeTransactionExtension())


class Request(BaseRequest):

    def copy(self, *args, **kwargs):
        """
        Copy the request and environment object.

        This only does a shallow copy, except of wsgi.input
        """
        self.make_body_seekable()
        env = self.environ.copy()
        new_req = self.__class__(env, *args, **kwargs)
        new_req.copy_body()
        new_req.identity = self.identity
        return new_req

    def get_authn_request(self):
        app = self.app.get_authn_provider(self)
        return self.copy(app=app)


class DBSessionRequest(Request):

    _db_session = None

    @property
    def db_session(self) -> sqlalchemy.orm.Session:
        if self._db_session is None:
            self._db_session = Session()
        return self._db_session


class BaseApp(CRUDApp, cors.CORSApp, SignalApp):

    authn_provider = dectate.directive(directive.AuthnProviderAction)
    request_class = Request

    def get_authn_provider(self, request):
        authn_app = self._get_authn_provider()
        authn_app.root = request.app.root
        authn_app.parent = request.app
        return authn_app

    @reg.dispatch_method()
    def _get_authn_provider(self):
        raise NotImplementedError

    def __repr__(self):
        return 'Morp Application -> %s:%s' % (self.__class__.__module__,
                                              self.__class__.__name__)


class SQLApp(TransactionApp, BaseApp):

    request_class = DBSessionRequest

    engine = None
    _raw_settings: dict = {}

    def __init__(self, engine=None, *args, **kwargs):
        super(SQLApp, self).__init__(*args, **kwargs)
        self.engine = engine
        self._init_engine()

    def _init_engine(self, session=Session) -> Engine:

        settings = self._raw_settings

        if self.engine is not None:
            return self.engine

        register_session(session)
        # initialize SQLAlchemy
        if not settings['application']['dburi']:
            raise ConfigurationError('SQLAlchemy settings not found')

        cwd = os.environ.get('MORP_WORKDIR', os.getcwd())
        os.chdir(cwd)
        dburi = settings['application']['dburi'] % {'here': cwd}
        engine = sqlalchemy.create_engine(dburi, poolclass=NullPool)
        session.configure(bind=engine)

        self.engine = engine
        return engine

    def initdb(self, session=Session):
        Base.metadata.create_all(self.engine)



from more.jwtauth import JWTIdentityPolicy
from morepath import Identity, NO_IDENTITY
import rulez
from .path import get_apikey_collection
from ..base import AuthnPolicy as BaseAuthnPolicy
from ...app import SQLApp
from .app import App as BaseAuthApp
from .user.model import UserSchema, UserModel
from .group.model import GroupSchema, GroupModel
from .apikey.model import APIKeySchema, APIKeyModel
from .storage.sqlstorage.sqlstorage import GroupSQLStorage, UserSQLStorage, APIKeySQLStorage
from .storage.memorystorage import GroupMemoryStorage, UserMemoryStorage, APIKeyMemoryStorage


class JWTWithAPIKeyIdentityPolicy(JWTIdentityPolicy):

    def identify(self, request):
        api_key = request.headers.get('X-API-KEY', None)
        if api_key:
            api_identity, api_secret = api_key.split('.')
            apikeys = get_apikey_collection(request)
            keys = apikeys.search(
                rulez.field['api_identity'] == api_identity, secure=False)
            if keys and keys[0].data['api_secret'] == api_secret:
                userid = keys[0].data['userid']
                return Identity(userid=userid)
        return super(JWTWithAPIKeyIdentityPolicy, self).identify(request)


class DefaultAuthnPolicy(BaseAuthnPolicy):

    def get_identity_policy(self, settings):
        jwtauth_settings = settings.jwtauth
        if jwtauth_settings:
            # Pass the settings dictionary to the identity policy.
            return JWTWithAPIKeyIdentityPolicy(**jwtauth_settings.__dict__.copy())
        raise Exception('JWTAuth configuration is not set')

    def verify_identity(self, app, identity):
        return True


class SQLStorageAuthApp(BaseAuthApp, SQLApp):
    pass


@SQLStorageAuthApp.storage(model=UserModel)
def get_user_sqlstorage(model, request, blobstorage):
    return UserSQLStorage(request, blobstorage=blobstorage)


@SQLStorageAuthApp.storage(model=GroupModel)
def get_group_sqlstorage(model, request, blobstorage):
    return GroupSQLStorage(request, blobstorage=blobstorage)


@SQLStorageAuthApp.storage(model=APIKeyModel)
def get_apikey_sqlstorage(model, request, blobstorage):
    return APIKeySQLStorage(request, blobstorage=blobstorage)


class SQLStorageAuthnPolicy(DefaultAuthnPolicy):

    app_cls = SQLStorageAuthApp


class MemoryStorageAuthApp(BaseAuthApp):
    pass


@MemoryStorageAuthApp.storage(model=UserModel)
def get_user_memorystorage(model, request, blobstorage):
    return UserMemoryStorage(request, blobstorage=blobstorage)


@MemoryStorageAuthApp.storage(model=GroupModel)
def get_group_memorystorage(model, request, blobstorage):
    return GroupMemoryStorage(request, blobstorage=blobstorage)


@MemoryStorageAuthApp.storage(model=APIKeyModel)
def get_apikey_memorystorage(model, request, blobstorage):
    return APIKeyMemoryStorage(request, blobstorage=blobstorage)


class MemoryStorageAuthnPolicy(DefaultAuthnPolicy):

    app_cls = MemoryStorageAuthApp



import re
import morepath
import dectate
from celery import Celery, Task
from celery import shared_task
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
#from typing import Union
from celery.local import Proxy
from celery.result import AsyncResult
from celery.schedules import crontab
from billiard.einfo import ExceptionInfo
from . import directive


class MorpTask(Task):
    abstract = True

    def after_return(self,
                     status: str,
                     retval: Any,
                     task_id: str,
                     args: List,
                     kwargs: Dict,
                     einfo: Optional[ExceptionInfo],
                     ):
        pass


class Signal(object):

    def __init__(self, app: morepath.App, signal: str, **kwargs):
        self.app = app
        self.signal = signal
        self.signal_opts = kwargs

    def subscribers(self) -> List[Proxy]:
        self.app.config.celery_subscriber_registry.setdefault(self.signal, [])
        return self.app.config.celery_subscriber_registry[self.signal]

    def send(self, request: morepath.Request, **kwargs) -> List[AsyncResult]:
        tasks = []
        envs = {}
        allcaps = re.compile(r'^[A-Z_]+$')
        for k, v in request.environ.items():
            if allcaps.match(k):
                envs[k] = v
            elif k in ['wsgi.url_scheme']:
                envs[k] = v

        req_json = {
            'headers': dict(request.headers),
            'environ': envs,
            'text': request.text,
        }

        subs = self.subscribers()
        for s in subs:
            wrapped = s.__wrapped__
            task = s.apply_async(kwargs=dict(request=req_json, **kwargs),
                                 **self.signal_opts)
            tasks.append(task)
        return tasks


class SignalApp(morepath.App):

    celery = Celery()
    _celery_subscribe = dectate.directive(directive.CelerySubscriberAction)

    @classmethod
    def async_subscribe(klass, signal: str, task_name: Optional[str] = None):
        def wrapper(wrapped):
            task = shared_task(name=task_name,  base=MorpTask)(wrapped)
            klass._celery_subscribe(signal)(task)
            return task
        return wrapper

    @classmethod
    def cron(klass, name: str, minute: str = '*', hour: str = '*', day_of_week: str = '*',
             day_of_month: str = '*', month_of_year: str = '*'):
        def wrapper(wrapped):
            task = shared_task()(wrapped)
            klass.celery.conf.beat_schedule[name] = {
                'task': '.'.join([wrapped.__module__, wrapped.__name__]),
                'schedule': crontab(minute=minute, hour=hour,
                                    day_of_week=day_of_week,
                                    day_of_month=day_of_month,
                                    month_of_year=month_of_year)
            }
            return task
        return wrapper

    def signal(self, signal: str, **kwargs) -> Signal:
        return Signal(self, signal, **kwargs)



from ...app import App
from ...user.model import UserSchema, UserModel
from ...group.model import GroupSchema, GroupModel
from ...apikey.model import APIKeyModel, APIKeySchema
from . import dbmodel as db
from morpfw.crud.storage.sqlstorage import SQLStorage
from morpfw.crud import errors as cruderrors
from ..interfaces import IGroupStorage, IUserStorage
import hashlib
import sqlalchemy as sa
from ... import exc


def hash(data):
    return hashlib.sha256(data.encode('utf-8')).hexdigest()


class UserSQLStorage(SQLStorage, IUserStorage):
    model = UserModel
    orm_model = db.User

    def create(self, data):
        data['password'] = hash(data['password'])
        return super(UserSQLStorage, self).create(data)

    def get_userid(self, model):
        return model.uuid

    def get_by_userid(self, userid, as_model=True):
        q = self.session.query(db.User).filter(db.User.uuid == userid)
        u = q.first()
        if not u:
            raise exc.UserDoesNotExistsError(userid)
        if not as_model:
            return u
        return self.model(self.request, self, u)

    def get_by_username(self, username, as_model=True):
        q = self.session.query(db.User).filter(db.User.username == username)
        u = q.first()
        if not u:
            raise exc.UserDoesNotExistsError(username)
        if not as_model:
            return u
        return self.model(self.request, self, u)

    def get_by_email(self, email):
        q = self.session.query(db.User).filter(db.User.email == email)
        u = q.first()
        if not u:
            return None
        return self.model(self.request, self, u)

    def change_password(self, userid, new_password):
        u = self.get_by_userid(userid)
        u.data['password'] = hash(new_password)

    def get_user_groups(self, userid):
        u = self.get_by_userid(userid, as_model=False)
        q = self.session.query(db.Membership).filter(
            db.Membership.user_id == u.id)
        membership = q.all()
        groupstorage = self.request.app.get_storage(GroupModel,
                                                    self.request)
        res = [groupstorage.get_by_id(m.group_id)
               for m in membership]
        return res

    def validate(self, userid, password):
        u = self.get_by_userid(userid)
        return u.data['password'] == hash(password)


class APIKeySQLStorage(SQLStorage):
    model = APIKeyModel
    orm_model = db.APIKey


class GroupSQLStorage(SQLStorage, IGroupStorage):
    model = GroupModel
    orm_model = db.Group

    def get_user_by_userid(self, userid, as_model=True):
        user_storage = self.app.get_storage(UserModel, self.request)
        return user_storage.get_by_userid(userid, as_model)

    def get_user_by_username(self, username, as_model=True):
        user_storage = self.app.get_storage(UserModel, self.request)
        return user_storage.get_by_username(username)

    def get_members(self, groupname):
        q = (self.session.query(db.User)
             .join(db.Membership)
             .join(db.Group)
             .filter(db.Group.groupname == groupname))
        members = []
        user_storage = self.app.get_storage(UserModel, self.request)
        for m in q.all():
            members.append(user_storage.model(self.request, user_storage, m))
        return members

    def add_group_members(self, groupname, userids):
        # FIXME: not using sqlalchemy relations might impact performance
        g = self.session.query(db.Group).filter(
            db.Group.groupname == groupname).first()
        if not g:
            raise ValueError("Group Does Not Exist %s" % groupname)
        gid = g.id
        for userid in userids:
            u = self.get_user_by_userid(userid, as_model=False)
            uid = u.id
            e = self.session.query(db.Membership).filter(
                sa.and_(db.Membership.group_id == gid,
                        db.Membership.user_id == uid)).first()
            if not e:
                m = db.Membership()
                m.group_id = gid
                m.user_id = uid
                self.session.add(m)

    def remove_group_members(self, groupname, userids):
        g = self.session.query(db.Group).filter(
            db.Group.groupname == groupname).first()
        if not g:
            raise exc.GroupDoesNotExistsError(groupname)
        gid = g.id
        for userid in userids:
            u = self.get_user_by_userid(userid, as_model=False)
            uid = u.id
            members = self.session.query(db.Membership).filter(
                sa.and_(db.Membership.group_id == gid,
                        db.Membership.user_id == uid)).all()
            for m in members:
                self.session.delete(m)

    def get_group_user_roles(self, groupname, userid):
        g = self.session.query(db.Group).filter(
            db.Group.groupname == groupname).first()
        if not g:
            raise exc.GroupDoesNotExistsError(groupname)
        gid = g.id
        u = self.get_user_by_userid(userid, as_model=False)
        uid = u.id
        roles = (self.session.query(db.RoleAssignment)
                 .join(db.Membership)
                 .filter(
            sa.and_(db.Membership.group_id == gid,
                    db.Membership.user_id == uid)).all())
        return [r.rolename for r in roles]

    def grant_group_user_role(self, groupname, userid, rolename):
        g = self.session.query(db.Group).filter(
            db.Group.groupname == groupname).first()
        if not g:
            raise exc.GroupDoesNotExistsError(groupname)
        gid = g.id
        u = self.get_user_by_userid(userid, as_model=False)
        uid = u.id
        m = self.session.query(db.Membership).filter(
            sa.and_(db.Membership.group_id == gid,
                    db.Membership.user_id == uid)).first()
        if not m:
            raise exc.MembershipError(userid, groupname)
        ra = self.session.query(db.RoleAssignment).filter(
            db.RoleAssignment.membership_id == m.id,
            db.RoleAssignment.rolename == rolename).first()
        if ra:
            return
        r = db.RoleAssignment()
        r.membership_id = m.id
        r.rolename = rolename
        self.session.add(r)

    def revoke_group_user_role(self, groupname, userid, rolename):
        g = self.session.query(db.Group).filter(
            db.Group.groupname == groupname).first()
        if not g:
            raise exc.GroupDoesNotExistsError(groupname)
        gid = g.id
        u = self.get_user_by_userid(userid, as_model=False)
        uid = u.id
        m = self.session.query(db.Membership).filter(
            sa.and_(db.Membership.group_id == gid,
                    db.Membership.user_id == uid)).first()
        if not m:
            raise exc.MembershipError(userid, groupname)
        ra = self.session.query(db.RoleAssignment).filter(
            db.RoleAssignment.membership_id == m.id,
            db.RoleAssignment.rolename == rolename).first()
        if ra:
            self.session.delete(ra)



from morepath import Identity, NO_IDENTITY
from .base import AuthnPolicy as BaseAuthnPolicy


class NoAuthIdentityPolicy(object):

    def identify(self, request):
        return Identity('ANONYMOUS')

    def remember(self, response, request, identity):
        pass

    def forget(self, response, request):
        pass


class AuthnPolicy(BaseAuthnPolicy):

    def get_identity_policy(self, settings):
        return NoAuthIdentityPolicy()

    def verify_identity(self, app, identity):
        return True



import reg
import inspect
import dectate
from . import pubsub
from uuid import uuid4

OBJECT_CREATED = 'object_created'
OBJECT_UPDATED = 'object_updated'
OBJECT_TOBEDELETED = 'object_tobedeleted'


def _get_identifier(obj):
    if inspect.isclass(obj):
        return [obj.__module__, obj.__name__]
    if getattr(obj, '__class__', None):
        return [obj.__module__, obj.__class__.__name__]
    raise ValueError("%s is neither instance nor class" % obj)


def construct_key(signal, app, obj):
    return ':'.join([signal] + _get_identifier(obj))


class Connect(dectate.Action):

    app_class_arg = True

    def __init__(self, model, signal):
        self.model = model
        self.signal = signal

    def identifier(self, app_class):
        return uuid4().hex

    def perform(self, obj, app_class):
        app_class._events.subscribe(
            obj,
            model=self.model, signal=self.signal)


class SignalApp(dectate.App):

    subscribe = dectate.directive(Connect)

    @reg.dispatch_method(
        reg.match_instance('model',
                           lambda self, request, obj, signal: obj),
        reg.match_key('signal',
                      lambda self, request, obj, signal: signal))
    def _events(self, request, obj, signal):
        raise NotImplementedError

    def signal_publish(self, request, obj, signal):
        return self._events.publish(self, request, obj, signal)



from morpfw.crud import Schema
from morpfw.crud.validator import regex_validator
from ..model import NAME_PATTERN, EMAIL_PATTERN
from ..app import App
from dataclasses import dataclass, field
import typing


@dataclass
class RegistrationSchema(object):
    # pattern=NAME_PATTERN)
    username: typing.Optional[str] = field(default=None, metadata={
        'morpfw': {
            'required': True,
            'validators': [regex_validator(NAME_PATTERN, 'name')]
        }
    })

    email: typing.Optional[str] = field(default=None, metadata={
        'morpfw': {
            'required': True,
            'validators': [regex_validator(EMAIL_PATTERN, 'email')]
        }
    })

    password: typing.Optional[str] = field(default=None, metadata={
        'morpfw': {'required': True}})
    password_validate: typing.Optional[str] = field(default=None, metadata={
        'morpfw': {'required': True}})


@dataclass
class LoginSchema(object):
    username: typing.Optional[str] = field(default=None, metadata={
        'morpfw': {'required': True}})

    password: typing.Optional[str] = field(default=None, metadata={
        'morpfw': {'required': True}})


@dataclass
class UserSchema(Schema):

    username: typing.Optional[str] = field(default=None, metadata={
        'morpfw': {
            'required': True,
            'validators': [regex_validator(NAME_PATTERN, 'name')]
        }
    })

    email: typing.Optional[str] = field(default=None, metadata={
        'morpfw': {
            'required': True,
            'validators': [regex_validator(EMAIL_PATTERN, 'email')]
        }
    })

    password: typing.Optional[str] = field(default=None)
    source: typing.Optional[str] = field(default='local')


@App.identifierfields(schema=UserSchema)
def user_identifierfields(schema):
    return ['username']



from .dictprovider import DictProvider
from ..app import App
from ..storage.elasticsearchstorage import ElasticSearchStorage
from ...interfaces import ISchema, IDataProvider


@App.dataprovider(schema=ISchema, obj=dict,
                  storage=ElasticSearchStorage)
def get_dataprovider(schema, obj, storage):
    return DictProvider(schema, obj, storage)



from .common import get_client
from morpfw.app import SQLApp
from more.basicauth import BasicAuthIdentityPolicy
import time
import pytest


class App(SQLApp):
    pass


class Root(object):
    pass


def get_identity_policy():
    return BasicAuthIdentityPolicy


def verify_identity(identity):
    return True


@App.path(model=Root, path='')
def get_root(request):
    return Root()


@App.json(model=Root)
def index(context, request):
    subs = request.app.signal(
        'test_signal').send(request, obj={'data': 10})
    cel = request.app.celery
    res = []
    for s in subs:
        try:
            res.append(s.get())
        except:
            pass
    return res


@App.async_subscribe('test_signal')
def handler1(request, obj):
    obj['handler'] = 'handler1'
    obj['data'] += 1
    return obj


@App.async_subscribe('test_signal')
def handler2(request, obj):
    obj['handler'] = 'handler2'
    obj['data'] += 5
    return obj


@App.async_subscribe('test_signal')
def handler3(request, obj):
    obj['handler'] = 'handler3'
    raise Exception('Error')


@pytest.fixture(scope='session')
def celery_config():
    return {
        'broker_url': 'amqp://guest:guest@localhost:34567/',
        'result_backend': 'db+postgresql://postgres@localhost:45678/morp_tests'
    }


def test_signal(pika_connection_channel, pgsql_db, celery_worker):
    c = get_client(App, get_identity_policy=get_identity_policy,
                   verify_identity=verify_identity)
    c.authorization = ('Basic', ('dummy', 'dummy'))

    r = c.get('/')

    res = list(sorted(r.json, key=lambda x: x['handler']))
    assert res[0]['data'] == 11
    assert res[1]['data'] == 15

#    r = c.get('/api/v1/task/+search?refresh=true')
#    res = list(sorted(r.json['results'], key=lambda x: x['data']['task']))
#    assert res[2]['data']['status'] == 'FAILURE'



import morpfw
from morpfw import sql as morpsql
from morpfw.app import BaseApp
from morpfw.crud.schema import Schema
from morpfw.authn.pas.app import App as AuthApp
from morpfw.authn.pas.user.path import get_user_collection
import sqlalchemy as sa
import jsl
from .test_auth import get_client
import morepath


class App(morpfw.SQLApp):
    pass


class MountedApp(AuthApp, morpfw.SQLApp):
    pass


class AppRoot(object):
    pass


@App.mount(app=MountedApp, path='/api/v1')
def mount_app(app):
    return MountedApp()


@App.json(model=AppRoot, name='login', request_method='POST')
def login(context, request):
    newreq = request.copy(app=MountedApp())
    coll = get_user_collection(newreq)
    data = newreq.json
    return {
        'authenticated': True if coll.authenticate(**data) else False
    }


@App.path(model=AppRoot, path='/')
def get_root():
    return AppRoot()


def test_morp_framework(pgsql_db):
    c = get_client(App, config='settings-mount.yml')

    r = c.post_json(
        '/api/v1/user/+login', {'username': 'admin',
                                'password': 'password'})

    c.authorization = ('JWT', r.headers.get('Authorization').split()[1])

    r = c.get('/api/v1/user/+refresh_token')

    c.authorization = ('JWT', r.headers.get('Authorization').split()[1])

    r = c.post_json('/+login', {'username': 'admin', 'password': 'password'})
    assert r.json['authenticated'] == True

    r = c.post_json('/+login', {'username': 'admin', 'password': 'invalid'})
    assert r.json['authenticated'] == False



import sqlalchemy as sa
import uuid
from sqlalchemy.ext.declarative import declarative_base
from rulez import compile_condition
from sqlalchemy.types import TypeDecorator, CHAR
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy import func
from datetime import datetime
from decimal import Decimal
import sqlalchemy_jsonfield as sajson
from .base import BaseStorage
from ..app import App
import jsl
import uuid


class SQLStorage(BaseStorage):

    _temp: dict = {}

    @property
    def orm_model(self):
        raise NotImplementedError

    @property
    def session(self):
        return self.request.db_session

    def create(self, data):
        o = self.orm_model()
        dst = self.app.get_dataprovider(self.model.schema, o, self)
        src = self.app.get_dataprovider(self.model.schema, data, self)
        for k, v in src.items():
            dst[k] = v
        m = self.model(self.request, self, o)
        identifier = m.identifier
        self.session.add(o)
        self.session.flush()
        self.session.refresh(o)
        return m

    def aggregate(self, query=None, group=None, order_by=None):
        group_bys = []

        if group:
            fields = []
            for k, v in group.items():
                if isinstance(v, str):
                    c = getattr(self.orm_model, v)
                    fields.append(c)
                    group_bys.append(c)
                elif isinstance(v, dict):
                    ff = v['function']
                    f = v['field']
                    c = getattr(self.orm_model, f)
                    if ff == 'count':
                        fields.append(func.count(c).label(k))
                    elif ff == 'sum':
                        fields.append(func.sum(c).label(k))
                    elif ff == 'avg':
                        fields.append(func.avg(c).label(k))
                    elif ff == 'year':
                        op = func.date_part('YEAR', c).label(k)
                        fields.append(op)
                        group_bys.append(op)
                    elif ff == 'month':
                        op = func.date_part('MONTH', c).label(k)
                        fields.append(op)
                        group_bys.append(op)
                    elif ff == 'day':
                        op = func.date_part('DAY', c).label(k)
                        fields.append(op)
                        group_bys.append(op)
                    else:
                        raise ValueError('Unknown function %s' % ff)
        else:
            fields = [self.orm_model]

        if query:
            f = compile_condition('sqlalchemy', query)
            filterquery = f(self.orm_model)
            filterquery = sa.and_(
                self.orm_model.deleted.is_(None), filterquery)
            q = self.session.query(*fields).filter(filterquery)
        else:
            q = self.session.query(
                *fields).filter(self.orm_model.deleted.is_(None))

        if order_by is not None:
            col = order_by[0]
            d = order_by[1]
            if d not in ['asc', 'desc']:
                raise KeyError(d)
            colattr = getattr(self.orm_model, col)
            if d == 'desc':
                q = q.order_by(colattr.desc())
            else:
                q = q.order_by(colattr)

        if group_bys:
            q = q.group_by(*group_bys)

        results = []
        for o in q.all():
            d = o._asdict()
            for k, v in d.items():
                if isinstance(v, datetime):
                    d[k] = v.isoformat()
                elif isinstance(v, Decimal):
                    d[k] = float(v)
                elif isinstance(v, uuid.UUID):
                    d[k] = v.hex

            results.append(d)
        return results

    def search(self, query=None, offset=None, limit=None, order_by=None):
        if query:
            f = compile_condition('sqlalchemy', query)
            filterquery = f(self.orm_model)
            q = self.session.query(self.orm_model).filter(sa.and_(
                self.orm_model.deleted.is_(None), filterquery))
        else:
            q = self.session.query(self.orm_model).filter(
                self.orm_model.deleted.is_(None))

        if order_by is not None:
            col = order_by[0]
            d = order_by[1]
            if d not in ['asc', 'desc']:
                raise KeyError(d)
            colattr = getattr(self.orm_model, col)
            if d == 'desc':
                q = q.order_by(colattr.desc())
            else:
                q = q.order_by(colattr)
        if offset is not None:
            q = q.offset(offset)
        if limit is not None:
            q = q.limit(limit)

        return [self.model(self.request, self, o) for o in q.all()]

    def get(self, identifier):
        qs = []
        for f, v in zip(
                self.app.get_identifierfields(self.model.schema),
                identifier.split(self.app.get_compositekey_separator())):
            qs.append(getattr(self.orm_model, f) == v)
        q = self.session.query(self.orm_model).filter(
            sa.and_(self.orm_model.deleted.is_(None), *qs))
        r = q.first()
        if not r:
            return None
        return self.model(self.request, self, r)

    def get_by_id(self, id):
        q = self.session.query(self.orm_model).filter(self.orm_model.id == id)
        r = q.first()
        if not r:
            return None
        return self.model(self.request, self, r)

    def get_by_uuid(self, uuid):
        uuid_field = self.app.get_uuidfield(self.model.schema)
        if getattr(self.orm_model, uuid_field, None) is None:
            raise AttributeError('%s does not have %s field' %
                                 (self.orm_model, uuid_field))
        qs = getattr(self.orm_model, uuid_field) == uuid
        q = self.session.query(self.orm_model).filter(qs)
        r = q.first()
        if not r:
            return None
        return self.model(self.request, self, r)

    def update(self, identifier, data):
        qs = []
        for f, v in zip(
                self.app.get_identifierfields(self.model.schema),
                identifier.split(
                    self.app.get_compositekey_separator())):
            qs.append(getattr(self.orm_model, f) == v)
        qs.append(self.orm_model.deleted.is_(None))
        q = self.session.query(self.orm_model).filter(sa.and_(*qs))

        r = q.first()
        if not r:
            raise ValueError(identifier)

        d = self.app.get_dataprovider(self.model.schema, r, self)
        for k, v in data.items():
            if d.get(k, None) != v:
                d[k] = v

        return self.model(self.request, self, r)

    def delete(self, identifier, model):
        model['deleted'] = datetime.utcnow()


class GUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    """
    impl = CHAR

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == 'postgresql':
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                # hexstring
                return "%.32x" % value.int

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            return uuid.UUID(value)


class BaseMixin(object):

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    uuid = sa.Column(GUID, default=uuid.uuid4, index=True, unique=True)
    created = sa.Column(sa.DateTime, default=datetime.utcnow)
    creator = sa.Column(sa.String(length=1024))
    modified = sa.Column(sa.DateTime, default=datetime.utcnow)
    state = sa.Column(sa.String(length=1024))
    deleted = sa.Column(sa.DateTime)
    blobs = sa.Column(sajson.JSONField)
    xattrs = sa.Column(sajson.JSONField)


Base = declarative_base(cls=BaseMixin)



import jsl
from morpfw.crud.storage.memorystorage import MemoryStorage
from morpfw.crud.blobstorage.fsblobstorage import FSBlobStorage
import morpfw.crud.signals as signals
from .crud_common import get_client, run_jslcrud_test, PageCollection, PageModel
from .crud_common import ObjectCollection, ObjectModel
from .crud_common import NamedObjectCollection, NamedObjectModel
from .crud_common import App as BaseApp
from .crud_common import BlobObjectCollection, BlobObjectModel
from .crud_common import FSBLOB_DIR
from more.transaction import TransactionApp
from more.basicauth import BasicAuthIdentityPolicy


class App(BaseApp):
    pass


class PageStorage(MemoryStorage):
    model = PageModel


@App.path(model=PageCollection, path='pages')
def collection_factory(request):
    storage = PageStorage(request)
    return PageCollection(request, storage)


@App.path(model=PageModel, path='pages/{identifier}')
def model_factory(request, identifier):
    storage = PageStorage(request)
    return storage.get(identifier)


class ObjectStorage(MemoryStorage):
    incremental_id = True
    incremental_column = 'id'
    model = ObjectModel


@App.path(model=ObjectCollection, path='objects')
def object_collection_factory(request):
    storage = ObjectStorage(request)
    return ObjectCollection(request, storage)


@App.path(model=ObjectModel, path='objects/{identifier}')
def object_model_factory(request, identifier):
    storage = ObjectStorage(request)
    return storage.get(identifier)


class NamedObjectStorage(MemoryStorage):
    model = NamedObjectModel


@App.path(model=NamedObjectCollection, path='named_objects')
def namedobject_collection_factory(request):
    storage = NamedObjectStorage(request)
    return NamedObjectCollection(request, storage)


@App.path(model=NamedObjectModel, path='named_objects/{identifier}')
def namedobject_model_factory(request, identifier):
    storage = NamedObjectStorage(request)
    o = storage.get(identifier)
    return o


class BlobObjectStorage(MemoryStorage):
    model = BlobObjectModel


@App.storage(model=BlobObjectModel)
def get_blobobject_storage(model, request, blobstorage):
    return BlobObjectStorage(request, blobstorage=blobstorage)


@App.blobstorage(model=BlobObjectModel)
def get_blobobject_blobstorage(model, request):
    return FSBlobStorage(request, FSBLOB_DIR)


@App.path(model=BlobObjectCollection, path='blob_objects')
def blobobject_collection_factory(request):
    storage = request.app.get_storage(BlobObjectModel, request)
    return BlobObjectCollection(request, storage)


@App.path(model=BlobObjectModel, path='blob_objects/{identifier}')
def blobobject_model_factory(request, identifier):
    storage = request.app.get_storage(BlobObjectModel, request)
    return storage.get(identifier)


def test_memorystorage():
    client = get_client(App)
    run_jslcrud_test(client, skip_aggregate=True)



import importlib
import reg
import morepath
from .app import SQLApp, Session, BaseApp
from .authn.pas.user.model import UserCollection, UserSchema, UserModel
from .authn.pas.group.model import GroupSchema, GroupModel
from .sql import Base
import os
from zope.sqlalchemy import register as register_session
from more.basicauth import BasicAuthIdentityPolicy
from .exc import ConfigurationError
import transaction
import sqlalchemy
from celery import Celery
import yaml
import copy

default_settings = yaml.load(
    open(os.path.join(os.path.dirname(__file__), 'default_settings.yml')))


@reg.dispatch(reg.match_class('app', lambda app, *args, **kwargs: app))
def create_app(app, settings, scan=True, **kwargs):
    raise NotImplementedError


@create_app.register(app=BaseApp)
def create_baseapp(app, settings, scan=True, **kwargs):

    s = copy.deepcopy(default_settings)
    for k in settings.keys():
        if k in s.keys():
            for j, v in settings[k].items():
                s[k][j] = v
        else:
            s[k] = settings[k]

    settings = s

    # initialize app

    if scan:
        morepath.autoscan()
        for scanmodpath in (settings['morpfw']['scan'] or []):
            scanmod = importlib.import_module(scanmodpath)
            morepath.scan(package=scanmod)

    app_settings = settings['application']

    authnpolicy_settings = app_settings['authn_policy_settings']
    authnpol_mod, authnpol_clsname = (
        app_settings['authn_policy'].strip().split(':'))

    authnpolicy = getattr(importlib.import_module(
        authnpol_mod), authnpol_clsname)(authnpolicy_settings)

    get_identity_policy = authnpolicy.get_identity_policy
    verify_identity = authnpolicy.verify_identity

    mounted_apps = app_settings['mounted_apps']
    if getattr(authnpolicy, 'app_cls', None):
        mounted_apps.append({
            'app_cls': authnpolicy.app_cls,
            'authn_policy': app_settings['authn_policy'],
            'authn_policy_settings': app_settings['authn_policy_settings']
        })
    for iapp in mounted_apps:
        if 'app_cls' in iapp.keys():
            iapp_cls = iapp['app_cls']
        else:
            iapp_path = iapp['app']
            iapp_mod, iapp_clsname = iapp_path.strip().split(':')
            iapp_cls = getattr(importlib.import_module(iapp_mod), iapp_clsname)

        if iapp.get('authn_policy', None):
            iapp_authnpolicy_settings = iapp.get('authn_policy_settings', {})
            iapp_authnpol_mod, iapp_authnpol_clsname = (
                iapp['authn_policy'].strip().split(':'))
            iapp_authnpolicy = getattr(importlib.import_module(
                iapp_authnpol_mod), iapp_authnpol_clsname)(iapp_authnpolicy_settings)
            iapp_get_identity_policy = iapp_authnpolicy.get_identity_policy
            iapp_verify_identity = iapp_authnpolicy.verify_identity
            iapp_cls.identity_policy()(iapp_get_identity_policy)
            iapp_cls.verify_identity()(iapp_verify_identity)
            if getattr(iapp_cls, 'authn_provider', None):
                iapp_cls.authn_provider()(iapp_authnpolicy.get_app)
        else:
            iapp_cls.identity_policy()(get_identity_policy)
            iapp_cls.verify_identity()(verify_identity)
            if getattr(iapp_cls, 'authn_provider', None):
                iapp_cls.authn_provider()(authnpolicy.get_app)

        iapp_cls.init_settings(settings)
        iapp_cls._raw_settings = settings
        iapp_cls.commit()

    app.identity_policy()(get_identity_policy)
    app.verify_identity()(verify_identity)
    app.authn_provider()(lambda: authnpolicy.app_cls())
    app.init_settings(settings)
    app._raw_settings = settings

    if settings['application']['development_mode']:
        os.environ['MOREPATH_TEMPLATE_AUTO_RELOAD'] = "1"

    app.commit()

    if settings['worker']['enabled']:
        celery_settings = settings['worker']['celery_settings']
        app.celery.conf.update(**celery_settings)
    application = app()
    return application


@create_app.register(app=SQLApp)
def create_sqlapp(app, settings, scan=True, **kwargs):

    application = create_baseapp(app=app, settings=settings, scan=scan,
                                 **kwargs)

    application.initdb()
    return application


def create_admin(app: morepath.App, username: str, password: str, email: str, session=Session):
    appreq = app.request_class(app=app, environ={'PATH_INFO': '/'})
    authapp = app.get_authn_provider(appreq)
    request = authapp.request_class(app=authapp, environ={'PATH_INFO': '/'})

    transaction.manager.begin()
    get_authn_storage = authapp.get_storage
    usercol = UserCollection(request, get_authn_storage(UserModel, request))
    userobj = usercol.create({'username': username,
                              'password': password,
                              'email': email,
                              'state': 'active'})
    gstorage = get_authn_storage(GroupModel, request)
    group = gstorage.get('__default__')
    group.add_members([userobj.userid])
    group.grant_member_role(userobj.userid, 'administrator')
    transaction.manager.commit()
    return userobj


def run(app, settings, host='127.0.0.1', port=5000, ignore_cli=True):
    application = create_app(app, settings)
    morepath.run(application, host=host, port=port, ignore_cli=ignore_cli)



from .app import App
from . import signals
from . import model
from datetime import datetime
from uuid import uuid4


@App.subscribe(signal=signals.OBJECT_CREATED, model=model.Model)
def set_uuid(app, request, obj, signal):
    uuid_field = app.get_uuidfield(obj.schema)
    if uuid_field in obj.schema.__dataclass_fields__.keys():
        if obj.data.get(uuid_field, None) is None:
            obj.data[uuid_field] = uuid4().hex


@App.subscribe(signal=signals.OBJECT_CREATED, model=model.Model)
def set_created(app, request, obj, signal):
    if 'created' in obj.schema.__dataclass_fields__.keys():
        now = datetime.utcnow()
        obj.data['created'] = now
        obj.data['modified'] = now


@App.subscribe(signal=signals.OBJECT_CREATED, model=model.Model)
def set_creator(app, request, obj, signal):
    if 'creator' in obj.schema.__dataclass_fields__.keys():
        obj.data['creator'] = request.identity.userid or ''


@App.subscribe(signal=signals.OBJECT_UPDATED, model=model.Model)
def set_modified(app, request, obj, signal):
    if 'modified' in obj.schema.__dataclass_fields__.keys():
        obj.data['modified'] = datetime.utcnow().isoformat()



import os
import morepath
import yaml
from webtest import TestApp as Client
from morpfw.crud.model import Collection, Model
from morpfw.crud.schema import Schema
from morpfw.crud import permission as crudperm
from morpfw.crud.model import StateMachine, XattrProvider
from morpfw.crud.xattrprovider import FieldXattrProvider
from morpfw.app import BaseApp
from morpfw.authn.base import AuthnPolicy as BaseAuthnPolicy
from morpfw.interfaces import ISchema
import morpfw.crud.signals as signals
import jsl
import json
from uuid import uuid4
from datetime import datetime
from morpfw.main import create_app
from more.basicauth import BasicAuthIdentityPolicy
import tempfile
import shutil
from dataclasses import dataclass, field
import typing

FSBLOB_DIR = tempfile.mkdtemp()


class App(BaseApp):
    pass


@App.permission_rule(model=Model, permission=crudperm.All)
def allow_all_model_access(identity, context, permission):
    return True


@App.permission_rule(model=Collection, permission=crudperm.All)
def allow_all_collection_access(identity, context, permission):
    return True


def validate_body(request, json):
    if not isinstance(json['body'], str):
        return

    if json['body'].lower() == "invalid":
        return "Body must not be 'invalid'"


@dataclass
class PageSchema(Schema):

    title: str = ''
    body: str = ''
    value: typing.Optional[int] = None
    footer: typing.Optional[str] = ''


@App.formvalidators(schema=PageSchema)
def page_formvalidators(schema):
    return [validate_body]


class PageCollection(Collection):
    schema = PageSchema


class PageModel(Model):
    schema = PageSchema


@dataclass
class ObjectSchema(Schema):

    body: str = ''
    created_flag: typing.Optional[bool] = False
    updated_flag: typing.Optional[bool] = False
    attrs: typing.Optional[dict] = field(default_factory=dict)


class ObjectCollection(Collection):
    schema = ObjectSchema


class ObjectModel(Model):
    schema = ObjectSchema


@App.subscribe(signal=signals.OBJECT_CREATED, model=ObjectModel)
def object_created(app, request, obj, signal):
    obj.data['created_flag'] = True


@App.subscribe(signal=signals.OBJECT_UPDATED, model=ObjectModel)
def object_updated(app, request, obj, signal):
    obj.data['updated_flag'] = True


@dataclass
class ObjectXattrSchema(ISchema):

    message: typing.Optional[str] = None


class ObjectXattrProvider(FieldXattrProvider):

    schema = ObjectXattrSchema


@App.xattrprovider(model=ObjectModel)
def get_objectmodel_xattrprovider(context):
    return ObjectXattrProvider(context)


class PageStateMachine(StateMachine):

    states = ['new', 'pending', 'approved']
    transitions = [
        {'trigger': 'approve', 'source': [
            'new', 'pending'], 'dest': 'approved'},
        {'trigger': 'submit', 'source': 'new', 'dest': 'pending'}
    ]


@App.statemachine(model=PageModel)
def get_pagemodel_statemachine(context):
    return PageStateMachine(context)


@dataclass
class NamedObjectSchema(Schema):

    name: typing.Optional[str] = None
    body: typing.Optional[str] = ''
    created_flag: typing.Optional[bool] = False
    updated_flag: typing.Optional[bool] = False


@App.identifierfields(schema=NamedObjectSchema)
def namedobject_identifierfields(schema):
    return ['name']


@App.default_identifier(schema=NamedObjectSchema)
def namedobject_default_identifier(schema, obj, request):
    return obj['name']


class NamedObjectCollection(Collection):
    schema = NamedObjectSchema


@App.json(model=NamedObjectCollection, name='get_uuid')
def get_object_by_uuid(context, request):
    uuid = request.GET.get('uuid')
    return context.get_by_uuid(uuid).json()


class NamedObjectModel(Model):
    schema = NamedObjectSchema


class BlobObjectSchema(Schema):
    pass


class BlobObjectCollection(Collection):
    schema = BlobObjectSchema


class BlobObjectModel(Model):
    schema = BlobObjectSchema

    blob_fields = ['blobfile']


class AuthnPolicy(BaseAuthnPolicy):

    def get_identity_policy(self, settings):
        return BasicAuthIdentityPolicy()

    def verify_identity(self, app, identity):
        if identity.userid == 'admin' and identity.password == 'admin':
            return True
        return False


def get_client(app, config='settings.yml'):
    if isinstance(config, str):
        with open(os.path.join(os.path.dirname(__file__), config)) as f:
            settings = yaml.load(f) or {}
    else:
        settings = config

    appobj = create_app(app, settings)
    c = Client(appobj)
    return c


def run_jslcrud_test(c, skip_aggregate=False):

    c.authorization = ('Basic', ('admin', 'admin'))

    # test loading the model details and schema
    r = c.get('/pages')

    assert r.json['schema']['type'] == 'object'

    # lets try creating an entry
    r = c.post_json('/pages/', {'title': 'Hello', 'body': 'World'})

    assert r.json['data']['title'] == 'Hello'

    uuid = r.json['data']['uuid']
    assert uuid
    assert len(uuid) == 32

    r = c.get('/pages/%s' % uuid)

    assert r.json['data']['title'] == 'Hello'
    assert r.json['data']['state'] == 'new'

    # lets see if the entry appears in listing

    for i in range(10):
        r = c.post_json(
            '/pages/', {'title': 'page%s' % i, 'body': 'page%sbody' % i, 'value': i})

    if not skip_aggregate:
        r = c.get('/pages/+aggregate', {'group': json.dumps({
            'count': {
                'function': 'count',
                'field': 'uuid'
            },
            'year': {
                'function': 'year',
                'field': 'created'
            },
            'month': {
                'function': 'month',
                'field': 'created'
            },
            'day': {
                'function': 'day',
                'field': 'created'
            },
            'sum': {
                'function': 'sum',
                'field': 'value'
            },
            'avg': {
                'function': 'avg',
                'field': 'value'
            }
        })})

        now = datetime.utcnow()
        assert r.json[0]['year'] == now.year
        assert r.json[0]['month'] == now.month
        assert r.json[0]['day'] == now.day
        assert r.json[0]['sum'] == 45
        assert r.json[0]['count'] == 11
        assert r.json[0]['avg'] == 4.5

    r = c.get('/pages/+search',
              {'q': json.dumps({'operator': 'in',
                                'value': ['Hello', 'something'],
                                'field': 'title'})})

    assert r.json['results'][0]['data']['title'] == 'Hello'

    r = c.get('/pages/+search')

    assert 'Hello' in [i['data']['title'] for i in r.json['results']]

    r = c.get('/pages/+search', {
        'select': '$.title',
        'q': json.dumps({'operator': 'in',
                         'value': ['Hello'],
                         'field': 'title'})
    })

    assert r.json['results'] == [['Hello']]

    r = c.get('/pages/+search', {
        'select': '$.[title, body]',
        'q': json.dumps({'operator': 'in',
                         'value': ['Hello'],
                         'field': 'title'})
    })

    assert r.json['results'] == [['Hello', 'World']]

    r = c.get('/pages/+search', {
        'order_by': 'title'
    })

    assert list([i['data']['title'] for i in r.json['results']]) == (
        ['Hello'] + ['page%s' % i for i in range(10)])

    r = c.get('/pages/+search', {
        'order_by': 'title:desc'
    })

    assert list([i['data']['title'] for i in r.json['results']]) == (
        ['page%s' % i for i in range(9, -1, -1)] + ['Hello'])

    r = c.get('/pages/+search', {
        'order_by': 'title',
        'limit': 5
    })

    assert list([i['data']['title'] for i in r.json['results']]) == (
        ['Hello'] + ['page%s' % i for i in range(4)])

    r = c.get('/pages/+search', {
        'order_by': 'title',
        'offset': 1
    })

    assert list([i['data']['title'] for i in r.json['results']]) == (
        ['page%s' % i for i in range(10)])

    r = c.get('/pages/+search', {
        'order_by': 'title',
        'offset': 1,
        'limit': 5
    })

    assert list([i['data']['title'] for i in r.json['results']]) == (
        ['page%s' % i for i in range(5)])

    # lets create another with wrong invalid values
    r = c.post_json('/pages/',
                    {'title': 'page2', 'body': 123, 'footer': 123},
                    expect_errors=True)

    assert r.json['status'] == 'error'
    assert len(r.json['field_errors']) == 2
    assert len(r.json['form_errors']) == 0

    r = c.post_json('/pages/',
                    {'title': 'page2', 'body': 'invalid', 'footer': 123},
                    expect_errors=True)

    assert r.json['status'] == 'error'
    assert len(r.json['field_errors']) == 1
    assert len(r.json['form_errors']) == 1

    # lets update the entry
    r = c.patch_json('/pages/%s' % uuid, {'body': 'newbody'})

    assert r.json['status'] == 'success'

    r = c.get('/pages/%s' % uuid)

    assert r.json['data']['body'] == 'newbody'

    # lets approve the page

    r = c.post_json('/pages/%s/+statemachine' %
                    uuid, {'transition': 'approve'})

    assert r.json['data']['state'] == 'approved'

    # it cant be approved twice

    r = c.post_json('/pages/%s/+statemachine' % uuid, {
        'transition': 'approve'
    }, expect_errors=True)

    assert r.status_code == 422

    r = c.patch_json('/pages/%s' %
                     uuid, {'body': 'invalid'}, expect_errors=True)

    assert r.json['status'] == 'error'

    # lets delete the entry

    r = c.get('/pages/+search')

    assert len(r.json['results']) == 11

    r = c.delete('/pages/%s' % uuid)

    assert r.json['status'] == 'success'

    r = c.get('/pages/%s' % uuid, expect_errors=True)

    r = c.get('/pages/+search')

    assert len(r.json['results']) == 10

    r = c.get('/objects')

    assert r.json

    r = c.post_json('/objects/', {'body': 'hello'})

    assert r.json['data']['body'] == 'hello'
    assert r.json['data']['created_flag'] is True
    assert r.json['data']['created']
    assert r.json['data']['creator'] == 'admin'

    obj_link = r.json['links'][0]['href']
    obj_xattr_link = obj_link + '/+xattr'
    r = c.get(obj_xattr_link)

    assert r.json == {}

    r = c.get(obj_link)

    assert r.json['data'].get('xattrs', None) is None

    r = c.get(obj_link + '/+xattr-schema')

    assert r.json['schema']['$schema']

    r = c.patch_json(obj_xattr_link, {'message': 'hello world'})

    r = c.get(obj_xattr_link)

    assert r.json == {'message': 'hello world'}

    r = c.get(obj_link)
    assert r.json['data']['xattrs'] == {'message': 'hello world'}

    # test creation of named object
    r = c.post_json('/named_objects/',
                    {'name': 'obj1', 'body': 'hello'})

    r = c.get('/named_objects/obj1')

    assert r.json['data']['name'] == 'obj1'
    uuid = r.json['data']['uuid']
    original_object = r.json['data']
    r = c.get('/named_objects/+get_uuid?uuid=%s' % uuid)
    object_by_uuid = r.json['data']
    assert original_object == object_by_uuid

    # duplicate should fail
    r = c.post_json('/named_objects/',
                    {'name': 'obj1', 'body': 'hello'},
                    expect_errors=True)

    assert r.status_code == 422

    # catch issue with ':' in name

    r = c.post_json('/named_objects/',
                    {'name': 'object:obj2', 'body': 'hello'})

    r = c.get('/named_objects/object:obj2')

    assert r.json['data']['name'] == 'object:obj2'

    r = c.get('/named_objects/object:obj2?select=$.[body]')

    assert r.json == ['hello']

    # catch issue with ' ' in name

    r = c.post_json('/named_objects/',
                    {'name': 'object obj2', 'body': 'hello'})

    r = c.get('/named_objects/object%20obj2')

    assert r.json['data']['name'] == 'object obj2'

    r = c.patch_json('/named_objects/object%20obj2', {'body': 'hello1'})

    assert r.status_code == 200

    # blob upload test

    r = c.post_json('/blob_objects', {})

    bloburl = r.json['links'][0]['href']

    r = c.get(bloburl)

    testimg = os.path.join(os.path.dirname(__file__), 'testimg.png')

    r = c.post(bloburl + '/+blobs?field=blobfile',
               upload_files=[('upload', testimg)])

    assert r.status_code == 200

    r = c.get(bloburl)

    assert r.json['data']['blobs']['blobfile']

    r = c.get(bloburl + '/+blobs?field=blobfile')

    with open(testimg, 'rb') as ti:
        assert r.body == ti.read()

    assert r.headers.get('Content-Type') == 'image/png'

    # unocide textfile upload test

    r = c.post_json('/blob_objects', {})

    bloburl = r.json['links'][0]['href']

    r = c.get(bloburl)

    testimg = os.path.join(os.path.dirname(__file__), 'testtxt.txt')

    r = c.post(bloburl + '/+blobs?field=blobfile',
               upload_files=[('upload', testimg)])

    assert r.status_code == 200

    r = c.get(bloburl)

    assert r.json['data']['blobs']['blobfile']

    r = c.get(bloburl + '/+blobs?field=blobfile')

    with open(testimg, 'rb') as ti:
        assert r.body == ti.read()

    assert r.headers.get('Content-Type') == 'text/plain; charset=UTF-8'

    shutil.rmtree(FSBLOB_DIR)



from .sqlstorage.sqlstorage import UserSQLStorage


class LDAP3SQLUserStorage(UserSQLStorage):

    def get(self, identifier):
        raise NotImplementedError

    def get_by_userid(self, userid):
        raise NotImplementedError

    def get_by_username(self, username):
        raise NotImplementedError

    def change_password(self, userid, new_password):
        raise NotImplementedError

    def validate(self, userid, password):
        raise NotImplementedError



from .group.path import get_group_collection


def rellink(context, request, name=None, method='GET', link_name=None):
    link_name = link_name or name
    if name:
        name = '+%s' % name
    if name is None:
        link_name = 'self'
    return {
        'rel': link_name,
        'type': method,
        'href': request.link(context, name)
    }


def has_role(request, role, userid=None, groupname='__default__'):
    group_col = get_group_collection(request)
    defgroup = group_col.get(groupname)
    if not userid:
        userid = request.identity.userid
    roles = defgroup.get_member_roles(userid)
    return role in roles



# FIXME: this test should run properly as unit test
from .common import get_client
from morpfw.app import SQLApp
import morpfw
import pytest
import morepath


class App(SQLApp):
    pass


@App.cron(name='test-cron')
def tick():
    print('tick!')


def main():
    morepath.autoscan()
    morepath.scan()
    App.commit()
    beat = App.celery.Beat()
    beat.run()


if __name__ == '__main__':
    import test_scheduler
    test_scheduler.main()



from ..app import App
from .model import GroupCollection, GroupModel
from .model import GroupSchema, MemberSchema
from ..validator import validate
from ..utils import rellink
from morpfw.util import get_user
from morpfw.crud import permission
from morpfw.crud.errors import UnprocessableError
from morpfw.crud.validator import validate_schema, get_data
from webob.exc import HTTPNotFound, HTTPForbidden, HTTPInternalServerError


@App.json(model=GroupModel,
          name='members',
          permission=permission.View)
def list_members(context, request):
    """Return the list of users in the group."""
    members = context.members()
    return {
        'users': [{
            'username': m.identifier,
            'userid': m.userid,
            'roles': context.get_member_roles(m.userid),
            'links': [rellink(m, request)]
        } for m in members]
    }


@App.json(model=GroupModel, name='grant', request_method='POST',
          permission=permission.Edit)
def grant_member(context, request):
    """Grant member roles in the group."""
    mapping = request.json['mapping']
    for entry in mapping:
        user = entry['user']
        roles = entry['roles']
        username = user.get('username', None)
        userid = user.get('userid', None)
        if userid:
            u = context.get_user_by_userid(userid)
        elif username:
            u = context.get_user_by_username(username)
        else:
            u = None
        if u is None:
            raise UnprocessableError(
                'User %s does not exists' % (userid or username))
        for rolename in roles:
            context.grant_member_role(u.userid, rolename)
    return {'status': 'success'}


@App.json(model=GroupModel, name='revoke', request_method='POST',
          permission=permission.Edit)
def revoke_member(context, request):
    """Revoke member roles in the group."""

    mapping = request.json['mapping']
    for entry in mapping:
        user = entry['user']
        roles = entry['roles']
        username = user.get('username', None)
        userid = user.get('userid', None)
        if userid:
            u = context.get_user_by_userid(userid)
        elif username:
            u = context.get_user_by_username(username)
        else:
            u = None
        if u is None:
            raise UnprocessableError(
                'User %s does not exists' % (userid or username))
        for rolename in roles:
            context.revoke_member_role(u.userid, rolename)
    return {'status': 'success'}



from ..model import NAME_PATTERN
from morpfw.crud import Collection, Model, Schema
from ..app import App
import jsl
from morpfw.crud import errors as cruderrors
from .. import exc
from .schema import GroupSchema, MemberSchema
from ..exc import GroupExistsError


class GroupCollection(Collection):
    schema = GroupSchema
    exist_exc = GroupExistsError


class GroupModel(Model):
    schema = GroupSchema

    def get_user_by_userid(self, userid):
        return self.storage.get_user_by_userid(userid)

    def get_user_by_username(self, username):
        return self.storage.get_user_by_username(username)

    def members(self):
        members = self.storage.get_members(self.identifier)
        active_members = [
            member for member in members if member.data.get('state') == 'active']
        return active_members

    def add_members(self, userids):
        self.storage.add_group_members(self.identifier, userids)

    def remove_members(self, userids):
        self.storage.remove_group_members(self.identifier, userids)

    def get_member_roles(self, userid):
        return self.storage.get_group_user_roles(self.identifier, userid)

    def grant_member_role(self, userid, rolename):
        if userid not in [m.userid for m in self.members()]:
            self.add_members([userid])
        self.storage.grant_group_user_role(self.identifier, userid, rolename)

    def revoke_member_role(self, userid, rolename):
        self.storage.revoke_group_user_role(
            self.identifier, userid, rolename)
        if (not self.get_member_roles(userid) and
                self.identifier != '__default__'):
            self.remove_members([userid])

    def _links(self):
        links = []
        for m in self.members():
            links.append({
                'rel': 'member',
                'href': self.request.link(m)
            })
        return links



from ..app import App
from ..types import datestr
from ..storage.memorystorage import MemoryStorage
from ...interfaces import ISchema, IDataProvider
from dateutil.parser import parse as parse_date
from ..util import dataclass_check_type, dataclass_get_type
import datetime
from dataclasses import _MISSING_TYPE

_MARKER: list = []


class DictProvider(IDataProvider):

    def __init__(self, schema, data, storage):
        self.schema = schema
        self.data = data
        self.storage = storage
        self.changed = False

    def __getitem__(self, key):
        t = dataclass_check_type(
            self.schema.__dataclass_fields__[key], datetime.datetime)
        if t:
            data = self.data[key]
            if isinstance(data, str):
                return parse_date(data)
            return data
        if key not in self.data.keys():
            field = self.schema.__dataclass_fields__[key]
            if field.default is not _MISSING_TYPE:
                return field.default
            if field.default_factory is not _MISSING_TYPE:
                return field.default_factory()
            return None
        return self.data[key]

    def __setitem__(self, key, value):
        field = self.schema.__dataclass_fields__[key]
        t = dataclass_get_type(field)
        for v in t['metadata']['validators']:
            v(value)
        if t['type'] == datetime.datetime:
            if value and not isinstance(value, datetime.datetime):
                value = parse_date(value)
        self.data[key] = value
        self.changed = True

    def __delitem__(self, key):
        del self.data[key]
        self.changed = True

    def setdefault(self, key, value):
        field = self.schema.__dataclass_fields__[key]
        t = dataclass_get_type(field)
        for v in t['metadata']['validators']:
            v(value)
        r = self.data.setdefault(key, value)
        self.changed = True
        return r

    def get(self, key, default=_MARKER):
        if default is _MARKER:
            return self.data.get(key)
        return self.data.get(key, default)

    def set(self, key, value):
        field = self.schema.__dataclass_fields__[key]
        t = dataclass_get_type(field)
        for v in t['metadata']['validators']:
            v(value)
        self.data[key] = value
        self.changed = True

    def items(self):
        return self.data.items()

    def keys(self):
        return self.data.keys()

    def as_dict(self):
        result = {}
        for k, v in obj.data.items():
            result[k] = v
        return result

    def as_json(self):
        result = {}
        for k, v in self.data.items():
            field = self.schema.__dataclass_fields__[k]
            exclude_if_empty = field.metadata.get(
                'morpfw', {}).get('exclude_if_empty', False)
            if exclude_if_empty and not v:
                continue
            if isinstance(v, datetime.datetime):
                result[k] = datestr(v.isoformat())
            else:
                result[k] = v
        return result


@App.dataprovider(schema=ISchema, obj=dict, storage=MemoryStorage)
def get_dataprovider(schema, obj, storage):
    return DictProvider(schema, obj, storage)


@App.jsonprovider(obj=DictProvider)
def get_jsonprovider(obj):
    return obj.as_json()



import dectate
import reg
import celery
from uuid import uuid4


class AuthnProviderAction(dectate.Action):

    app_class_arg = True

    def identifier(self, app_class):
        return str((app_class, uuid4().hex))

    def perform(self, obj, app_class):
        app_class._get_authn_provider.register(reg.methodify(obj))



from functools import partial
from .errors import ValidationError, FormValidationError, FieldValidationError
from jsonschema import Draft4Validator
from .util import jsl_to_jsonobject, dataclass_to_jsl, dataclass_get_type
import reg
from morepath.publish import resolve_model
import urllib
import re


@reg.dispatch(reg.match_instance('model'), reg.match_instance('request'))
def get_data(model, request):
    raise NotImplementedError


def regex_validator(pattern, name):
    p = re.compile(pattern)

    def _regex_validator(value):
        if not p.match(value):
            raise FieldValidationError(
                '%s does not match %s pattern' % (value, name))

    return _regex_validator


def load(validator, schema, request):
    newreq = request.app.request_class(
        request.environ.copy(), request.app.root,
        path_info=urllib.parse.unquote(request.path))
    context = resolve_model(newreq)
    context.request = request
    if schema is None:
        dc = context.schema
    else:
        dc = schema
    jslschema = dataclass_to_jsl(dc, nullable=True)
    schema = jslschema.get_schema(ordered=True)
    form_validators = request.app.get_formvalidators(dc)
    params = {}

    validator.check_schema(schema)
    v = validator(schema)
    data = get_data(context, request)
    field_errors = sorted(v.iter_errors(data), key=lambda e: e.path)
    if field_errors:
        params['field_errors'] = field_errors
    form_errors = []
    for form_validator in form_validators:
        e = form_validator(request, request.json)
        if e:
            form_errors.append(FormValidationError(e))

    if form_errors:
        params['form_errors'] = form_errors

    if params:
        raise ValidationError(**params)

    # field errors
    for k, f in dc.__dataclass_fields__.items():
        t = dataclass_get_type(f)
        for validate in t['metadata']['validators']:
            validate(data[k])
    return request.json


def validate_schema(validator=Draft4Validator, schema=None):
    return partial(load, validator, schema)



from .model import UserCollection, UserModel, UserSchema
from ..app import App


def get_user(request, identifier):
    collection = get_user_collection(request)
    return collection.get(identifier)


def get_user_collection(request):
    authprovider = request.app.get_authn_provider(request)
    return UserCollection(request, storage=authprovider.get_storage(UserModel, request))


@App.path(model=UserModel,
          path='user/{username}',
          variables=lambda obj: {
              'username': obj.data['username']
          })
def _get_user(app, request, username):
    return get_user(request, username)


@App.path(model=UserCollection,
          path='user')
def _get_user_collection(app, request):
    return get_user_collection(request)



import morepath
from .app import App
from . import subscribers
from .app import Session
import argparse
import yaml
import sqlalchemy
import os
from .model import Collection, Model
from .rulesadapter import Adapter
from .schema import Schema
from .model import StateMachine
from .util import resolve_model
from .app import App
from .storage.sqlstorage import SQLStorage
from zope.sqlalchemy import register as register_session


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--settings', default='settings.yml')
    args = parser.parse_args()
    with open(args.settings) as cf:
        settings = yaml.load(cf)

    application = create_app(App, settings)

    # start app
    morepath.run(application)


def create_app(app, settings, sqlalchemy_session=Session,
               sqlalchemy_bases=None):
    sqlalchemy_bases = sqlalchemy_bases or []
    register_session(sqlalchemy_session)
    # initialize SQLAlchemy
    if 'sqlalchemy' in settings:
        cwd = os.getcwd()
        engine = sqlalchemy.create_engine(
            settings['sqlalchemy']['dburi'] % {'here': cwd})
        sqlalchemy_session.configure(bind=engine)

    # initialize app
    app.init_settings(settings)
    morepath.commit(app)
    morepath.autoscan()
    app.commit()
    application = app()

    # create tables
    if 'sqlalchemy' in settings:
        for base in sqlalchemy_bases:
            base.metadata.create_all(engine)

    return application


if __name__ == '__main__':
    run()



from ..app import App
from .model import APIKeyModel, APIKeyCollection
from ..path import get_user, get_user_collection
from ..path import get_apikey_collection, get_apikey
from morpfw.crud import permission
import uuid
import rulez
import morepath


@App.json(model=APIKeyCollection, request_method='POST',
          permission=permission.Create)
def generate_apikey(context, request):
    usercol = get_user_collection(request)
    user = usercol.get_by_userid(request.identity.userid)
    data = {}
    if not user.validate(request.json['password']):
        @request.after
        def adjust_response(response):
            response.status = 422

        return {'status': 'error', 'message': 'Invalid password'}

    data['label'] = request.json['label']
    obj = context.create(data)
    return obj.json()



from .base import BaseStorage
from rulez import compile_condition
import elasticsearch.exceptions as es_exc
from ..app import App
from pprint import pprint as print
import copy
from typing import Optional


class AggGroup(object):

    def __init__(self, key, field, type='terms', opts=None, children=None):
        self.key = key
        self.field = field
        self.type = type
        self.children = children or []
        self.parent = None
        self.is_leaf = False
        self.opts = opts or {}

    def parse(self, result, parents=None):
        parents = parents or {}
        records = result[self.key]['buckets']
        results = []
        for r in records:
            key = r['key_as_string']
            if self.type == 'date_histogram':
                if self.opts.get('format', None) in ['yyyy', 'MM', 'dd']:
                    key = int(key)
            parents[self.key] = key
            if self.is_leaf:
                values = copy.copy(parents)
                for c in self.children:
                    if c.function == 'count':
                        values[c.key] = r['doc_count']
                    else:
                        values.update(c.parse(r, parents=parents))
                results.append(values)
            else:
                values = []
                for c in self.children:
                    values += c.parse(r, parents=parents)
                results += values
        return results

    def json(self):
        res = {
            self.type: {
                'field': self.field
            }
        }

        if self.opts:
            res[self.type].update(self.opts)

        if self.children:
            for c in self.children:
                res.setdefault('aggs', {})
                res['aggs'].update(c.json())

        return {
            self.key: res
        }


class AggValue(object):

    def __init__(self, key, function, field):
        self.key = key
        self.function = function
        self.field = field
        self.parent = None

    def parse(self, result, parents=None):
        if self.function == 'count':
            return {}
        parents = parents or {}
        p = copy.copy(parents)
        p[self.key] = result[self.key][self.field]
        return p

    def json(self):
        if self.function == 'count':
            return {}
        return {
            self.key: {
                self.function: {
                    'field': self.field
                }
            }
        }


class Aggregate(object):

    def __init__(self):
        self.groups = []
        self.values = []
        self.first_group = None
        self._finalized = False

    def add_group(self, key, field, type='terms', opts=None):
        if self._finalized:
            raise ValueError('Aggregate has been finalized')
        self.groups.append(AggGroup(key, field, type, opts))

    def add(self, key, function, field):
        if self._finalized:
            raise ValueError('Aggregate has been finalized')
        self.values.append(AggValue(key, function, field))

    def finalize(self):
        if self._finalized:
            raise ValueError('Aggregate has been finalized')

        prev = None
        first = None
        for g in self.groups:
            if prev is None:
                prev = g
                continue
            if g not in prev.children:
                prev.children.append(g)
                g.parent = prev
            prev = g
        prev.is_leaf = True
        for v in self.values:
            if v not in prev.children:
                prev.children.append(v)
                v.parent = prev

        self._finalized = True

    def json(self):
        if not self._finalized:
            self.finalize()
        return self.groups[0].json()

    def parse(self, result):
        if not self._finalized:
            self.finalize()
        return self.groups[0].parse(result)


class ElasticSearchStorage(BaseStorage):

    refresh: Optional[str] = None
    auto_id = False
    use_transactions = False

    @property
    def index_name(self):
        raise NotImplementedError

    @property
    def doc_type(self):
        raise NotImplementedError

    @property
    def client(self):
        return self.request.es_client

    def create_index(self):
        if not self.client.indices.exists(self.index_name):
            try:
                self.client.indices.create(index=self.index_name, body={
                    'settings': {
                        'number_of_shards': 1,
                        'number_of_replicas': 0
                    }
                })
            except es_exc.TransportError as e:
                if e.error == 'index_already_exists_exception':
                    return
                raise e

    def create(self, data):
        m = self.model(self.request, self, data)
        self.create_index()
        try:
            r = self.client.index(index=self.index_name,
                                  doc_type=self.doc_type,
                                  id=m.identifier,
                                  body=data, refresh=self.refresh)
        except es_exc.TransportError as e:
            print(e.args, e.error, e.info)
            print(data)
            raise e

        if self.auto_id:
            self.set_identifier(m.data, r['_id'])
            m.save()
        return m

    def search(self, query=None, offset=None, limit=None, order_by=None):
        if limit is None:
            limit = 9999
        if query:
            q = {'query': compile_condition('elasticsearch', query)()}
        else:
            q = {'query': {'match_all': {}}}
        if limit is not None:
            q['from'] = 0
            q['size'] = limit

        self.create_index()
        params = {}
        if offset is not None:
            params['from_'] = offset
        if limit:
            params['size'] = limit
        if order_by:
            params['sort'] = [':'.join(order_by)]
        res = self.client.search(index=self.index_name, doc_type=self.doc_type,
                                 body=q, **params)

        data = [self.model(self.request, self, o['_source'])
                for o in res['hits']['hits']]

        return list(data)

    def aggregate(self, query=None, group=None, order_by=None):
        if query:
            q = {'query': compile_condition('elasticsearch', query)()}
        else:
            q = {'query': {'match_all': {}}}

        q['size'] = 0

        self.create_index()
        params = {}
        if order_by:
            params['sort'] = [':'.join(order_by)]

        if group:
            aggs = Aggregate()

            for k, v in group.items():
                if isinstance(v, str):
                    aggs.add_group(k, v)

                elif isinstance(v, dict):
                    ff = v['function']
                    f = v['field']
                    if ff == 'count':
                        aggs.add(k, 'count', f)
                    elif ff == 'sum':
                        aggs.add(k, 'sum', f)
                    elif ff == 'avg':
                        aggs.add(k, 'avg', f)
                    elif ff == 'year':
                        aggs.add_group(k, f, type='date_histogram', opts={
                            'interval': 'year',
                            'format': 'yyyy'
                        })
                    elif ff == 'month':
                        aggs.add_group(k, f, type='date_histogram', opts={
                            'interval': 'month',
                            'format': 'MM'
                        })
                    elif ff == 'day':
                        aggs.add_group(k, f, type='date_histogram', opts={
                            'interval': 'day',
                            'format': 'dd'
                        })
                    elif ff == 'interval-1m':
                        aggs.add_group(k, f, type='date_histogram', opts={
                            'interval': '1m',
                            'format': "yyyy-MM-dd'T'HH:mm"
                        })
                    elif ff == 'interval-15m':
                        aggs.add_group(k, f, type='date_histogram', opts={
                            'interval': '15m',
                            'format': "yyyy-MM-dd'T'HH:mm"
                        })
                    elif ff == 'interval-30m':
                        aggs.add_group(k, f, type='date_histogram', opts={
                            'interval': '30m',
                            'format': "yyyy-MM-dd'T'HH:mm"
                        })
                    elif ff == 'interval-1h':
                        aggs.add_group(k, f, type='date_histogram', opts={
                            'interval': '1h',
                            'format': "yyyy-MM-dd'T'HH:mm"
                        })
                    else:
                        raise ValueError('Unknown function %s' % ff)

            aggs.finalize()
            q['aggs'] = aggs.json()
        res = self.client.search(index=self.index_name, doc_type=self.doc_type,
                                 body=q, **params)
        data = aggs.parse(res['aggregations'])

        return list(data)

    def get(self, identifier):
        self.create_index()
        try:
            res = self.client.get(index=self.index_name,
                                  doc_type=self.doc_type, id=identifier,
                                  refresh=self.refresh)
        except es_exc.NotFoundError as e:
            return None
        return self.model(self.request, self, res['_source'])

    def get_by_uuid(self, uuid):
        res = self.search({'field': 'uuid', 'operator': '==', 'value': uuid})
        if res:
            return self.model(self.request, self, res[0].json()['data'])
        return None

    def get_by_id(self, id):
        res = self.client.get(index=self.index_name, doc_type=self.doc_type,
                              id=id)
        return self.model(self.request, self, res['_source'])

    def update(self, identifier, data):
        self.create_index()
        self.client.update(index=self.index_name, doc_type=self.doc_type,
                           id=identifier, body={'doc': data},
                           refresh=self.refresh)

    def delete(self, identifier, model):
        self.create_index()
        self.client.delete(index=self.index_name, doc_type=self.doc_type,
                           id=identifier, refresh=True)



from .app import App
from .user.model import UserCollection, UserModel
from .user.path import get_user_collection, get_user
from .group.model import GroupModel, GroupCollection, GroupSchema
from .group.path import get_group, get_group_collection
from .apikey.model import APIKeyCollection, APIKeyModel, APIKeySchema
from .apikey.path import get_apikey, get_apikey_collection



import morpfw
import sqlalchemy as sa
from morpfw.crud import permission as crudperm
import typing
from .app import App


class PageSchema(morpfw.Schema):
    body: typing.Optional[str] = None


class PageCollection(morpfw.Collection):
    schema = PageSchema


class PageModel(morpfw.Model):
    schema = PageSchema


class Page(morpfw.SQLBase):
    __tablename__ = 'test_page'

    body = sa.Column(sa.String(length=1024))


class PageStorage(morpfw.SQLStorage):
    model = PageModel
    orm_model = Page


@App.path(model=PageCollection, path='pages')
def get_collection(request):
    storage = PageStorage(request)
    return PageCollection(request, storage)


@App.path(model=PageModel, path='pages/{identifier}')
def get_model(request, identifier):
    collection = get_collection(request)
    return collection.get(identifier)


@App.permission_rule(model=PageModel, permission=crudperm.All)
def allow_model(identity, context, permission):
    return True


@App.permission_rule(model=PageCollection, permission=crudperm.All)
def allow_collection(identity, context, permission):
    return True



from .authn.pas.group.model import GroupSchema, GroupCollection
from .authn.pas.path import get_group_collection, get_group
from .authn.pas.path import get_user_collection as get_user_collection


def create_group(request, groupname):
    collection = get_group_collection(request)
    return collection.create({'groupname': groupname})


def delete_group(request, groupname):
    collection = get_group_collection(request)
    group = collection.get(groupname)
    group.delete()


def get_user(request, username):
    collection = get_user_collection(request)
    return collection.get(username)



#
from . import crud
from .crud import Collection, Model, Adapter, Schema
from .crud import StateMachine
from .crud.storage.sqlstorage import SQLStorage
from .crud.blobstorage.fsblobstorage import FSBlobStorage
from .crud.storage.elasticsearchstorage import ElasticSearchStorage
from .sql import Base as SQLBase
from .app import SQLApp
from .main import create_app, run
from .main import create_admin
from .crud import signals as crudsignals
from .util import get_group, get_user



import morepath
from more.jsonschema import JsonSchemaApp
from . import signals as signals
from sqlalchemy.orm import sessionmaker
import dectate
import reg
from more.signals import SignalApp
from . import component as actions
import warnings
from .model import Model
from .blobstorage.base import NullBlobStorage

Session = sessionmaker()


class App(JsonSchemaApp, signals.SignalApp):
    @classmethod
    def jslcrud_dataprovider(klass, *args, **kwargs):
        warnings.warn(
            "jslcrud_dataprovider is deprecated, use dataprovider", DeprecationWarning)
        return klass.dataprovider(*args, **kwargs)

    dataprovider = dectate.directive(actions.DataProviderAction)

    @classmethod
    def jslcrud_jsonprovider(klass, *args, **kwargs):
        warnings.warn(
            "jslcrud_jsonprovider is deprecated, use jsonprovider", DeprecationWarning)
        return klass.jsonprovider(*args, **kwargs)

    jsonprovider = dectate.directive(actions.JSONProviderAction)

    @classmethod
    def jslcrud_formvalidators(klass, *args, **kwargs):
        warnings.warn(
            "jslcrud_formvalidators is deprecated, use formvalidators", DeprecationWarning)
        return klass.formvalidators(*args, **kwargs)

    formvalidators = dectate.directive(actions.FormValidatorAction)

    @classmethod
    def jslcrud_identifierfields(klass, *args, **kwargs):
        warnings.warn(
            "jslcrud_identifierfields is deprecated, use identifierfields", DeprecationWarning)
        return klass.identifierfields(*args, **kwargs)

    identifierfields = dectate.directive(actions.IdentifierFieldsAction)

    @classmethod
    def jslcrud_default_identifier(klass, *args, **kwargs):
        warnings.warn(
            "jslcrud_default_identifier is deprecated, use default_identifier", DeprecationWarning)
        return klass.default_identifier(*args, **kwargs)

    default_identifier = dectate.directive(actions.DefaultIdentifierAction)

    @classmethod
    def jslcrud_rulesadapter(klass, *args, **kwargs):
        warnings.warn(
            "jslcrud_rulesadapter is deprecated, use rulesadapter", DeprecationWarning)
        return klass.rulesadapter(*args, **kwargs)

    rulesadapter = dectate.directive(actions.RulesAdapterAction)

    @classmethod
    def jslcrud_uuidfield(klass, *args, **kwargs):
        warnings.warn(
            "jslcrud_uuidfield is deprecated, use uuidfield", DeprecationWarning)
        return klass.uuidfield(*args, **kwargs)

    uuidfield = dectate.directive(actions.UUIDFieldAction)

    @classmethod
    def jslcrud_statemachine(klass, *args, **kwargs):
        warnings.warn(
            "jslcrud_statemachine is deprecated, use statemachine", DeprecationWarning)
        return klass.statemachine(*args, **kwargs)

    statemachine = dectate.directive(actions.StateMachineAction)

    xattrprovider = dectate.directive(actions.XattrProviderAction)

    storage = dectate.directive(actions.StorageAction)

    blobstorage = dectate.directive(actions.BlobStorageAction)

    def get_storage(self, model, request):
        blobstorage = self.get_blobstorage(model, request)
        return self._get_storage(model, request, blobstorage)

    def get_blobstorage(self, model, request):
        return self._get_blobstorage(model, request)

    @reg.dispatch_method(reg.match_class('model'),
                         reg.match_instance('request'),
                         reg.match_instance('blobstorage'))
    def _get_storage(self, model, request, blobstorage):
        raise NotImplementedError

    @reg.dispatch_method(reg.match_class('model'), reg.match_instance('request'))
    def _get_blobstorage(self, model, request):
        return NullBlobStorage()

    @reg.dispatch_method(
        reg.match_class('schema',
                        lambda self, schema, obj, storage: schema),
        reg.match_instance('obj'),
        reg.match_instance('storage'))
    def get_dataprovider(self, schema, obj, storage):
        raise NotImplementedError('Dataprovider for %s/%s' % (
            storage.__class__, obj.__class__))

    @reg.dispatch_method(reg.match_instance('obj'))
    def get_jsonprovider(self, obj):
        raise NotImplementedError('JSONProvider for %s' % obj.__class__)

    @reg.dispatch_method(reg.match_class('schema',
                                         lambda self, schema: schema))
    def get_formvalidators(self, schema):
        return []

    @reg.dispatch_method(reg.match_class('schema',
                                         lambda self, schema: schema))
    def get_identifierfields(self, schema):
        raise NotImplementedError('IdentifierFields for %s' % schema)

    @reg.dispatch_method(reg.match_class('schema',
                                         lambda self, schema: schema))
    def get_uuidfield(self, schema):
        return 'uuid'

    @reg.dispatch_method(
        reg.match_class('schema',
                        lambda self, schema, obj, request: schema))
    def get_default_identifier(self, schema, obj, request):
        return None

    @reg.dispatch_method(reg.match_instance('model', lambda self, obj: obj))
    def get_rulesadapter(self, obj):
        raise NotImplementedError

    @reg.dispatch_method(reg.match_instance('model',
                                            lambda self, context: context))
    def get_statemachine(self, context):
        raise NotImplementedError

    @reg.dispatch_method(reg.match_instance('model',
                                            lambda self, context: context))
    def get_xattrprovider(self, context):
        raise NotImplementedError

    def get_compositekey_separator(self):
        morp_settings = self.settings.application
        return morp_settings.compositekey_separator

    def join_identifier(self, *args):
        separator = self.get_compositekey_separator()
        return separator.join(args)

    def permits(self, request: morepath.Request,
                context: Model, permission: str):
        identity = request.identity
        return self._permits(identity, context, permission)



import abc
import morepath
import webob
from typing import Optional, Union, BinaryIO, List, Sequence, Type


class ISchema(object):
    """
    Schema represent the fields available in a content type. MorpFW uses
    python dataclass as its schema we provide transformation function to
    transform dataclass into other schema format (eg: JSL, colander, etc)
    """
    pass


class IDataProvider(abc.ABC):
    """
    Data provider wraps around backend storage model object 
    (eg: SQLAlchemy model object) and provide a bridge for getting and setting
    data from model.
    """
    @abc.abstractmethod
    def __init__(self, schema: Type[ISchema], data: dict,
                 storage: 'IStorageBase'):
        super().__init__()

    @abc.abstractmethod
    def __setitem__(self, key, value):
        raise NotImplementedError

    @abc.abstractmethod
    def __getitem__(self, key):
        raise NotImplementedError

    @abc.abstractmethod
    def __delitem__(self, key):
        raise NotImplementedError

    @abc.abstractmethod
    def setdefault(self, key, value):
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, key, default=None):
        raise NotImplementedError

    @abc.abstractmethod
    def set(self, key, value):
        raise NotImplementedError

    @abc.abstractmethod
    def items(self):
        raise NotImplementedError

    @abc.abstractmethod
    def keys(self):
        raise NotImplementedError

    @abc.abstractmethod
    def as_dict(self):
        """Returns dictionary representation of the data, where python objects 
        such as ``datetime`` remains as python objects"""
        raise NotImplementedError

    @abc.abstractmethod
    def as_json(self):
        """Returns JSON-safe dictionary representation of the data. Any python
        objects are serialized into JSON-safe data type"""
        raise NotImplementedError


class IBlob(abc.ABC):

    @abc.abstractmethod
    def __init__(self, uuid, filename: str,
                 mimetype: Optional[str] = None,
                 size: Optional[int] = None,
                 encoding: Optional[str] = None):
        super().__init__()

    @abc.abstractmethod
    def open(self) -> BinaryIO:
        raise NotImplementedError

    @abc.abstractmethod
    def get_size(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def __call__(self, req: webob.Request) -> webob.Response:
        raise NotImplementedError


class IBlobStorage(abc.ABC):

    @abc.abstractmethod
    def put(self, field: str, fileobj: BinaryIO,
            filename: str,
            mimetype: Optional[str] = None,
            size: Optional[int] = None,
            encoding: Optional[str] = None,
            uuid: Optional[str] = None) -> IBlob:
        """Receive and store blob data"""
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, uuid: str) -> Optional[IBlob]:
        """Return blob data"""
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, uuid: str):
        """Delete blob data"""
        raise NotImplementedError


class IStorageBase(abc.ABC):
    """Basic storage backend"""

    @abc.abstractmethod
    def __init__(self, request: morepath.Request,
                 blobstorage: Optional[IBlobStorage] = None):
        super().__init__()

    @abc.abstractmethod
    def create(self, data: dict) -> 'IModel':
        """Create a model from submitted data"""
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, identifier) -> Optional['IModel']:
        """return model from identifier"""
        raise NotImplementedError

    @abc.abstractmethod
    def search(self, query: Optional[dict] = None, offset: Optional[int] = None,
               limit: Optional[int] = None,
               order_by: Union[None, list, tuple] = None) -> Sequence['IModel']:
        """return search result based on specified rulez query"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_by_id(self, id) -> Optional['IModel']:
        """return model from internal ID"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_by_uuid(self, uuid) -> Optional['IModel']:
        """return model from uuid"""
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, identifier, data):
        """update model with values from data"""
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, identifier, model):
        """delete model data"""
        raise NotImplementedError


class IStorage(IStorageBase):
    """Aggregateable storage"""

    @abc.abstractmethod
    def aggregate(self, query: Optional[dict] = None,
                  group: Optional[dict] = None,
                  order_by: Union[None, list, tuple] = None) -> list:
        """return aggregation result based on specified rulez query and group"""
        raise NotImplementedError


class IModel(abc.ABC):
    """Model is a representation of a data object. It 
    provide a common set of API which is then delegated down
    to the storage provider.

    Model is subscriptable and you can use it like a dictionary
    to access stored data.

    :param request: the request object
    :param storage: storage provider
    :param data: initial data on this model

    """

    #: Set whether object is linkable or not.
    #: If an object is linkable, its json result
    #: will have links attribute
    linkable: bool

    #: The dataclass schema which this model will be using
    schema: Type[ISchema]

    #: When set to True, will enable PATCH view to update model
    update_view_enabled: bool

    #: When set to True, will enable DELETE view to delete model
    delete_view_enabled: bool

    #: When set to True, will enable ``+statemachine`` view for state management
    statemachine_view_enabled: bool

    #: Field name on the model data which will be storing blob references
    blobstorage_field: str

    #: List of blob field names allowed on this model
    blob_fields: List[str]

    #: url identifier for this model
    identifier: str

    #: uuid of this model
    uuid: str

    #: Data provider
    data: IDataProvider

    @abc.abstractmethod
    def __init__(self, request: morepath.Request,
                 storage: IStorage, data: dict):
        super().__init__()

    @abc.abstractmethod
    def __setitem__(self, key, value):
        raise NotImplementedError

    @abc.abstractmethod
    def __getitem__(self, key):
        raise NotImplementedError

    @abc.abstractmethod
    def __delitem__(self, key):
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, newdata: dict):
        """Update model with new data"""
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self):
        """Delete model"""
        raise NotImplementedError

    @abc.abstractmethod
    def save(self):
        """Persist model data into backend storage"""
        raise NotImplementedError

    @abc.abstractmethod
    def json(self) -> dict:
        """Convert model to JSON-safe dictionary"""
        raise NotImplementedError

    @abc.abstractmethod
    def links(self) -> list:
        """Generate links for this model"""
        raise NotImplementedError

    @abc.abstractmethod
    def rules_adapter(self):
        """Return pluggable business rule adapter for this model"""
        raise NotImplementedError

    @abc.abstractmethod
    def statemachine(self):
        """Return PyTransition statemachine adapter for this model"""
        raise NotImplementedError

    @abc.abstractmethod
    def set_initial_state(self):
        """Initialize default statemachine state for this model"""
        raise NotImplementedError

    @abc.abstractmethod
    def put_blob(self, field: str, fileobj: BinaryIO,
                 filename: str,
                 mimetype: Optional[str] = None,
                 size: Optional[int] = None,
                 encoding: Optional[str] = None) -> IBlob:
        """Receive and store blob object"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_blob(self, field: str) -> IBlob:
        """Return blob"""
        raise NotImplementedError

    @abc.abstractmethod
    def delete_blob(self, field: str):
        """Delete blob"""
        raise NotImplementedError


class ICollection(abc.ABC):
    """Collection provide an API for querying group of model from its storage"""

    @abc.abstractmethod
    def search(self, query: Optional[dict] = None,
               offset: int = 0,
               limit: Optional[int] = None,
               order_by: Optional[tuple] = None,
               secure: bool = False) -> List[IModel]:
        """Search for models

        Filtering is done through ``rulez`` based JSON/dict query, which 
        defines boolean statements in JSON/dict structure.

        :param query: Rulez based query
        :param offset: Result offset
        :param limit: Maximum number of result
        :param order_by: Tuple of ``(field, order)`` where ``order`` is 
                         ``'asc'`` or ``'desc'``
        :param secure: When set to True, this will filter out any object which
                       current logged in user is not allowed to see

        :todo: ``order_by`` need to allow multiple field ordering
        """
        raise NotImplementedError

    @abc.abstractmethod
    def aggregate(self, query: Optional[dict] = None,
                  group: Optional[dict] = None,
                  order_by: Optional[tuple] = None) -> List[IModel]:
        """Get aggregated results

        :param query: Rulez based query
        :param group: Grouping structure
        :param order_by: Tuple of ``(field, order)`` where ``order`` is 
                         ``'asc'`` or ``'desc'``

        :todo: Grouping structure need to be documented

        """
        raise NotImplementedError

    @abc.abstractmethod
    def create(self, data: dict) -> IModel:
        """Create a model from data"""
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, identifier) -> IModel:
        """Get model by url identifier key"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_by_uuid(self, uuid: str) -> IModel:
        """Get model by uuid"""
        raise NotImplementedError

    @abc.abstractmethod
    def json(self) -> dict:
        """JSON-safe dictionary representing this collection"""
        raise NotImplementedError

    @abc.abstractmethod
    def links(self) -> list:
        """Links related to this collection"""
        raise NotImplementedError



import morepath
from morpfw.crud import Collection, Model, StateMachine
from morpfw.crud import errors as cruderrors
from ..app import App
import jsl
from .. import exc
import sqlalchemy as sa
import sqlalchemy_jsonfield as sajson
from ..model import NAME_PATTERN, EMAIL_PATTERN
from morpfw.crud import signals as crudsignal
from morpfw.crud import errors as cruderrors
from morpfw.crud.schema import Schema
from morpfw.crud.rulesadapter import Adapter
from morpfw.crud.validator import regex_validator
from ..group.model import GroupCollection, GroupSchema, GroupModel
from uuid import uuid4
import re
from .schema import RegistrationSchema, UserSchema, LoginSchema
from ..exc import UserExistsError
from ..utils import has_role


class UserCollection(Collection):
    schema = UserSchema
    exist_exc = UserExistsError

    def authenticate(self, username, password):
        if re.match(EMAIL_PATTERN, username):
            user = self.storage.get_by_email(username)
        else:
            user = self.storage.get(username)

        if user is None:
            return None
        if not user.validate(password):
            return None
        return user

    def get_by_userid(self, userid):
        return self.storage.get_by_userid(userid)

    def _create(self, data):
        return super(UserCollection, self)._create(data)


class UserModel(Model):

    schema = UserSchema

    blob_fields = ['profile-photo']

    @property
    def userid(self):
        return self.storage.get_userid(self)

    @property
    def identity(self):
        return morepath.Identity(self.userid)

    def change_password(self, password, new_password):
        if not has_role(self.request, 'administrator'):
            if not self.validate(password, check_state=False):
                raise exc.InvalidPasswordError(self.userid)
        self.storage.change_password(self.identity.userid, new_password)

    def validate(self, password, check_state=True):
        if check_state and self.data['state'] != 'active':
            return False
        return self.storage.validate(self.userid, password)

    def groups(self):
        return self.storage.get_user_groups(self.userid)

    def group_roles(self):
        group_roles = {}
        for g in self.groups():
            group_roles[g.data['groupname']] = g.get_member_roles(
                self.userid)
        return group_roles

    def _links(self):
        links = super()._links()
        links.append({
            'rel': 'userid',
            'value': self.userid
        })
        for g in self.groups():
            links.append({
                'rel': 'group',
                'name': g.identifier,
                'href': self.request.link(g),
            })
        return links


class UserStateMachine(StateMachine):

    states = ['active', 'inactive', 'deleted']
    transitions = [
        {'trigger': 'activate', 'source': 'inactive', 'dest': 'active'},
        {'trigger': 'deactivate', 'source': 'active', 'dest': 'inactive'},
    ]


@App.statemachine(model=UserModel)
def userstatemachine(context):
    return UserStateMachine(context)


class UserRulesAdapter(Adapter):

    def transform_json(self, data):
        data = data.copy()
        for f in ['password', 'password_validate']:
            if f in data.keys():
                del data[f]
        return data


@App.rulesadapter(model=UserModel)
def get_rulesadapter(obj):
    return UserRulesAdapter(obj)


@App.subscribe(signal=crudsignal.OBJECT_CREATED, model=UserModel)
def add_user_to_default_group(app, request, obj, signal):
    request = obj.request
    storage = app.get_storage(GroupModel, request)
    g = storage.get('__default__')
    if g is None:
        gcol = GroupCollection(request, storage)
        g = gcol.create({'groupname': '__default__'})
    g.add_members([obj.userid])



import morepath
from ..authn.pas.user.model import UserCollection, UserModel
from ..authn.pas.group.model import GroupCollection, GroupModel
from ..authn.pas.group.schema import GroupSchema
from ..authn.pas.apikey.model import APIKeyCollection, APIKeyModel
from ..authn.pas.group.path import get_group_collection
from ..crud.model import Model, Collection
from ..authn.pas import permission as authperm
from ..crud import permission as crudperm
from ..authn.pas.utils import has_role


class DefaultAuthzPolicy(morepath.App):
    pass


def _has_admin_role(context):
    return has_role(context.request, 'administrator')


@DefaultAuthzPolicy.permission_rule(model=Model, permission=crudperm.All)
def reject_all_model_access(identity, context, permission):
    return False


@DefaultAuthzPolicy.permission_rule(model=Collection, permission=crudperm.All)
def reject_all_collection_access(identity, context, permission):
    return False


@DefaultAuthzPolicy.permission_rule(model=UserCollection,
                                    permission=authperm.Register)
def allow_register(identity, context, permission):
    if _has_admin_role(context):
        return True
    return False


@DefaultAuthzPolicy.permission_rule(model=UserCollection,
                                    permission=crudperm.All)
def allow_users_management(identity, context, permission):
    if _has_admin_role(context):
        return True
    return False


@DefaultAuthzPolicy.permission_rule(model=GroupCollection,
                                    permission=crudperm.All)
def allow_groups_management(identity, context, permission):
    if _has_admin_role(context):
        return True
    return False


@DefaultAuthzPolicy.permission_rule(model=GroupModel, permission=crudperm.All)
def allow_group_management(identity, context, permission):
    if _has_admin_role(context):
        return True
    return False


@DefaultAuthzPolicy.permission_rule(model=UserModel, permission=crudperm.All)
def allow_user_management(identity, context, permission):
    if _has_admin_role(context):
        return True
    if context['uuid'] == identity.userid:
        return True
    return False


@DefaultAuthzPolicy.permission_rule(model=UserModel,
                                    permission=authperm.ChangePassword)
def allow_change_password(identity, context, permission):
    if _has_admin_role(context):
        return True
    if context['uuid'] == identity.userid:
        return True
    return False


@DefaultAuthzPolicy.permission_rule(model=APIKeyCollection, permission=crudperm.All)
def allow_apikeys_management(identity, context, permission):
    return True


@DefaultAuthzPolicy.permission_rule(model=APIKeyModel, permission=crudperm.All)
def allow_apikey_management(identity, context, permission):
    if _has_admin_role(context):
        return True
    if context['userid'] == identity.userid:
        return True
    return False



from pytest_postgresql import factories
import socket
from contextlib import closing
from pytest_rabbitmq.factories import rabbitmq_proc
import pytest
import pika
import os
import pytest
import shutil
import mirakuru
from elasticsearch import Elasticsearch
from pytest_postgresql import factories

pgsql_proc = factories.postgresql_proc(
    executable='/usr/bin/pg_ctl', host='localhost', port=45678,
    user='postgres')
pgsql_db = factories.postgresql('pgsql_proc', db='morp_tests')

rabbitmq_pika_proc = rabbitmq_proc(
    server='/usr/lib/rabbitmq/bin/rabbitmq-server',
    port=34567)


@pytest.fixture
def pika_connection_channel(rabbitmq_pika_proc):
    connection = pika.BlockingConnection(
        pika.ConnectionParameters(host='localhost', port=34567,
                                  credentials=pika.PlainCredentials('guest',
                                                                    'guest')))
    channel = connection.channel()
    return [connection, channel]


@pytest.fixture(scope='session')
def es_proc(request):
    port = 9085
    home_dir = '/tmp/elasticsearch_%s' % port
    os.environ['ES_HOME'] = home_dir
    command = [
        os.environ['ELASTICSEARCH_EXECUTABLE'],
        '-p', '/tmp/elasticsearch.%s.pid' % port,
        '-E', 'http.port=%s' % port,
        '-E', 'default.path.logs=/tmp/elasticsearch_%s_logs' % port,
        '-E', 'cluster.name=elasticsearch_cluster_%s' % port,
        '-E', "network.publish_host='127.0.0.1'",
        '-E', 'index.store.type=mmapfs'
    ]
    es_proc = mirakuru.HTTPExecutor(
        command, shell=True, url='http://127.0.0.1:%s' % port)
    es_proc.start()

    def finalize_elasticsearch():
        es_proc.stop()
        if os.path.exists(home_dir):
            shutil.rmtree(home_dir)

    request.addfinalizer(finalize_elasticsearch)
    return es_proc


@pytest.fixture(scope='session')
def es_client(request):
    process = request.getfixturevalue('es_proc')
    if not process.running():
        process.start()

    hosts = "%s:%s" % (process.host, process.port)

    client = Elasticsearch(hosts=hosts)

    def drop_indexes():
        client.indices.delete(index='*')

    request.addfinalizer(drop_indexes)

    return client



from .app import App
from .model import Model, Collection
from .validator import validate_schema, get_data
from .errors import ValidationError, StateUpdateProhibitedError
from .errors import AlreadyExistsError, UnprocessableError
from .errors import FieldValidationError
from . import permission
import json
from webob.exc import HTTPNotFound, HTTPForbidden, HTTPInternalServerError
from morepath.request import Request
from transitions import MachineError
from jsonpath_ng import parse as jsonpath_parse
from urllib.parse import urlencode
import traceback
import os
import sys
import logging
logger = logging.getLogger('morp')


@get_data.register(model=Collection, request=Request)
def get_collection_data(model, request):
    return request.json


@App.json(model=Collection, permission=permission.View)
def schema(context, request):
    return context.json()


@App.json(model=Collection, name='aggregate', permission=permission.Aggregate)
def aggregate(context, request):
    if not context.aggregate_view_enabled:
        raise HTTPNotFound()

    query = json.loads(request.GET.get('q', '{}'))
    if not query:
        query = None

    order_by = request.GET.get('order_by', None)
    group = json.loads(request.GET.get('group', '{}'))
    if order_by:
        order_by = order_by.split(':')
        if len(order_by) == 1:
            order_by = order_by + ['asc']
    objs = context.aggregate(query, group=group, order_by=order_by)
    return objs


@App.json(model=Collection, name='search', permission=permission.Search)
def search(context, request):
    if not context.search_view_enabled:
        raise HTTPNotFound()
    query = json.loads(request.GET.get('q', '{}'))
    if not query:
        query = None
    limit = int(request.GET.get('limit', 0)) or None
    offset = int(request.GET.get('offset', 0))
    order_by = request.GET.get('order_by', None)
    select = request.GET.get('select', None)
    if order_by:
        order_by = order_by.split(':')
        if len(order_by) == 1:
            order_by = order_by + ['asc']
    # HACK: +1 to ensure next page links is triggered
    searchlimit = limit
    if limit:
        searchlimit = limit + 1
    objs = context.search(query, offset=offset,
                          limit=searchlimit,
                          order_by=order_by,
                          secure=True)
    # and limit back to actual limit
    objs = [obj.json() for obj in objs]
    has_next = False
    if limit:
        if len(objs) > limit:
            has_next = True
        objs = objs[:limit]
    if select:
        expr = jsonpath_parse(select)
        results = []
        for obj in objs:
            results.append([match.value for match in expr.find(obj['data'])])
    else:
        results = objs
    params = {}
    if select:
        params['select'] = select
    if limit:
        params['limit'] = limit
    params['offset'] = offset + (limit or 0)
    qs = urlencode(params)
    res = {'results': results, 'q': query}
    if has_next:
        res.setdefault('links', [])
        res['links'].append({
            'rel': 'next',
            'href': request.link(context, '+search?%s' % qs)
        })
    if offset > 0:
        prev_offset = offset - (limit or 0)
        if prev_offset < 0:
            prev_offset = 0
        params['offset'] = prev_offset
        qs = urlencode(params)
        res.setdefault('links', [])
        res['links'].append({
            'rel': 'previous',
            'href': request.link(context, '+search?%s' % qs)
        })
    return res


@App.json(model=Collection, request_method='POST',
          load=validate_schema(), permission=permission.Create)
def create(context, request, json):
    if not context.create_view_enabled:
        raise HTTPNotFound()
    obj = context.create(request.json)
    obj.save()
    return obj.json()


@get_data.register(model=Model, request=Request)
def get_obj_data(model, request):
    data = model.json()['data']
    data.update(request.json)
    return data


@App.json(model=Model, permission=permission.View)
def read(context, request):
    select = request.GET.get('select', None)
    obj = context.json()
    if select:
        expr = jsonpath_parse(select)
        obj = [match.value for match in expr.find(obj['data'])]
    return obj


@App.json(model=Model, request_method='PATCH',
          load=validate_schema(),
          permission=permission.Edit)
def update(context, request, json):
    if not context.update_view_enabled:
        raise HTTPNotFound()

    context.update(request.json, secure=True)
    return {'status': 'success'}


@App.json(model=Model, name='statemachine', request_method='POST',
          permission=permission.Edit)
def statemachine(context, request):
    if not context.statemachine_view_enabled:
        raise HTTPNotFound()

    sm = context.state_machine()
    transition = request.json['transition']
    attr = getattr(sm, transition, None)
    if attr is None:
        @request.after
        def adjust_status(response):
            response.status = 422

        return {
            'status': 'error',
            'message': 'Unknown transition %s' % transition
        }
    attr()
    context.save()
    return context.json()


@App.json(model=Model, request_method='DELETE',
          permission=permission.Delete)
def delete(context, request):
    if not context.delete_view_enabled:
        raise HTTPNotFound()

    context.delete()
    return {'status': 'success'}


@App.view(model=Model, name='blobs', permission=permission.View)
def get_blob(context, request):
    field = request.GET.get('field', None)
    if field is None or context.storage.blobstorage is None:
        raise HTTPNotFound()

    blob = context.get_blob(field)

    return request.get_response(blob)


@App.json(model=Model, name='blobs', request_method='POST', permission=permission.Edit)
def put_blob(context, request):
    field = request.GET.get('field', None)
    if field is None or context.storage.blobstorage is None:
        raise HTTPNotFound()

    upload = request.POST.get('upload')

    if upload is None:
        raise UnprocessableError()

    context.put_blob(field, filename=os.path.basename(
        upload.filename), mimetype=upload.type,
        fileobj=upload.file)
    return context['blobs']


@App.json(model=Model, name='xattr-schema', permission=permission.View)
def get_xattr_schema(context, request):
    if not context.xattr_view_enabled:
        raise HTTPNotFound()

    provider = context.xattrprovider()
    return provider.jsonschema()


@App.json(model=Model, name='xattr', permission=permission.View)
def get_xattr(context, request):
    if not context.xattr_view_enabled:
        raise HTTPNotFound()

    provider = context.xattrprovider()
    return provider.as_json()


@App.json(model=Model, name='xattr', permission=permission.View,
          request_method='PATCH')
def set_xattr(context, request):
    if not context.xattr_view_enabled:
        raise HTTPNotFound()

    provider = context.xattrprovider()
    provider.process_update(request.json)
    return {'status': 'success'}


@App.json(model=AlreadyExistsError)
def alreadyexists_error(context, request):
    @request.after
    def adjust_status(response):
        response.status = 422

#   FIXME: should log this when a config for DEBUG_SECURITY is enabled
#    logger.error(traceback.format_exc())
#    print(traceback.format_exc())

    return {'status': 'error',
            'message': 'Object Already Exists : %s' % context.message}


@App.json(model=HTTPForbidden)
def forbidden_error(context, request):
    @request.after
    def adjust_status(response):
        response.status = 403
#   FIXME: should log this when a config for DEBUG_SECURITY is enabled
#    logger.error(traceback.format_exc())
    return {'status': 'error',
            'message': 'Access Denied : %s' % request.path}


@App.json(model=HTTPNotFound)
def httpnotfound_error(context, request):
    @request.after
    def adjust_status(response):
        response.status = 404
    return {'status': 'error',
            'message': 'Object Not Found : %s' % request.path}


@App.json(model=ValidationError)
def validation_error(context, request):
    @request.after
    def adjust_status(response):
        response.status = 422

    field_errors = []
    form_errors = []
    for e in context.field_errors:
        path = [str(p) for p in e.path]
        field_errors.append({
            'field': '.'.join(path),
            'message': e.message
        })

    for e in context.form_errors:
        form_errors.append(e.message)

    return {
        'status': 'error',
        'field_errors': field_errors,
        'form_errors': form_errors
    }


@App.json(model=FieldValidationError)
def field_validation_error(context, request):
    @request.after
    def adjust_status(response):
        response.status = 422

    form_errors = [context.args[0]]

    return {
        'status': 'error',
        'field_errors': [],
        'form_errors': form_errors
    }


@App.json(model=MachineError)
def statemachine_error(context, request):
    @request.after
    def adjust_status(response):
        response.status = 422

    return {
        'status': 'error',
        'message': context.value
    }


@App.json(model=StateUpdateProhibitedError)
def stateupdateprohibited_error(context, request):
    @request.after
    def adjust_status(response):
        response.status = 422

    return {
        'status': 'error',
        'message': "Please use 'statemachine' view to update state"
    }


@App.json(model=Exception)
def internalserver_error(context, request):
    @request.after
    def adjust_status(response):
        response.status = 500

    logger.error(traceback.format_exc())

    return {
        'status': 'error',
        'message': "Internal server error"
    }


@App.json(model=UnprocessableError)
def unprocessable_error(context, request):
    @request.after
    def adjust_status(response):
        response.status = 422
    return {'status': 'error',
            'message': '%s' % (context.message)}



import yaml
import os
from .test_auth import _test_authentication, get_client
from more.jwtauth import JWTIdentityPolicy
from more.transaction import TransactionApp
from morepath.reify import reify
from morepath.request import Request
from morpfw.app import SQLApp
from morpfw.authn.pas.app import App
from morpfw.authn.pas.policy import SQLStorageAuthnPolicy
from morpfw.authn.pas.policy import SQLStorageAuthApp as BaseAuthApp
from morpfw.authz.pas import DefaultAuthzPolicy


class SQLStorageAuthApp(BaseAuthApp, DefaultAuthzPolicy):
    pass


class SQLStorageApp(SQLApp, DefaultAuthzPolicy):
    pass


class AuthnPolicy(SQLStorageAuthnPolicy):
    app_cls = SQLStorageAuthApp


@SQLStorageApp.mount(app=AuthnPolicy.app_cls, path='/auth')
def mount_app(app):
    return AuthnPolicy.app_cls()


def test_authentication_sqlstorage(pgsql_db):
    with open(os.path.join(os.path.dirname(__file__),
                           'settings-sqlalchemy.yml')) as f:
        settings = yaml.load(f)

    c = get_client(SQLStorageApp, settings)

    _test_authentication(c)



from .model import GroupModel, GroupCollection, GroupSchema
from ..app import App


def get_group(request, identifier):
    collection = get_group_collection(request)
    return collection.get(identifier)


def get_group_collection(request):
    authnzprovider = request.app.get_authn_provider(request)
    storage = authnzprovider.get_storage(GroupModel, request)
    return GroupCollection(request, storage)


@App.path(model=GroupModel,
          path='group/{identifier}')
def _get_group(app, request, identifier):
    return get_group(request, identifier)


@App.path(model=GroupCollection,
          path='group')
def _get_group_collection(app, request):
    return get_group_collection(request)



from ..app import App
from .model import UserCollection, UserModel
from ..path import get_user_collection
from .model import UserSchema, LoginSchema
from .model import RegistrationSchema
from ..exc import UserExistsError
import morepath
from ..validator import validate
from ..utils import rellink, has_role
from .. import permission
from webob.exc import HTTPNotFound
from more.jwtauth import (
    verify_refresh_request, InvalidTokenError, ExpiredSignatureError
)
from morpfw.crud.util import dataclass_to_jsl
from morpfw.crud.validator import validate_schema
from morpfw.crud.errors import AlreadyExistsError


@App.json(model=UserCollection, name='register', request_method='POST',
          permission=permission.Register, load=validate_schema(schema=RegistrationSchema))
def register(context, request, load):
    """Validate the username and password and create the user."""
    data = request.json
    res = validate(data, dataclass_to_jsl(
        RegistrationSchema).get_schema())
    if res:
        @request.after
        def set_error(response):
            response.status = 422
        return {
            'status': 'error',
            'field_errors': [{'message': res[x]} for x in res.keys()]
        }

    if data['password'] != data['password_validate']:
        @request.after
        def adjust_response(response):
            response.status = 422

        return {'status': 'error',
                'message': 'Password confirmation does not match'}

    if 'state' not in data.keys() or not data['state']:
        data['state'] = request.app.settings.application.new_user_state
    del data['password_validate']
    obj = context.create(data)
    return {'status': 'success'}


@App.json(model=UserCollection, name='login')
def login(context, request):
    return {
        'schema': dataclass_to_jsl(LoginSchema).get_schema(),
        'links': [rellink(context, request, 'login', 'POST')]
    }


@App.json(model=UserCollection, name='create', request_method='POST')
def reject_create(context, request):
    raise HTTPNotFound()


@App.json(model=UserCollection, name='login', request_method='POST')
def process_login(context, request):
    """Authenticate username and password and log in user"""
    username = request.json['username']
    password = request.json['password']

    # Do the password validation.
    user = context.authenticate(username, password)
    if not user:
        @request.after
        def adjust_status(response):
            response.status = 401
        return {
            'status': 'error',
            'error': {
                'code': 401,
                'message': 'Invalid Username / Password'
            }
        }

    @request.after
    def remember(response):
        """Remember the identity of the user logged in."""
        # We pass the extra info to the identity object.
        response.headers.add('Access-Control-Expose-Headers', 'Authorization')
        identity = user.identity
        request.app.remember_identity(response, request, identity)

    return {
        'status': 'success'
    }


@App.json(model=UserCollection, name='refresh_token')
def refresh_token(context, request):
    try:
        # Verifies if we're allowed to refresh the token.
        # In this case returns the userid.
        # If not raises exceptions based on InvalidTokenError.
        # If expired this is a ExpiredSignatureError.
        userid = verify_refresh_request(request)
    except ExpiredSignatureError:
        @request.after
        def expired_nonce_or_token(response):
            response.status_code = 403
        return {
            'status': 'error',
            'message': 'Your session has expired.'
        }
    except InvalidTokenError as e:
        @request.after
        def invalid_token(response):
            response.status_code = 403
        return {
            'status': 'error',
            'message': 'Could not refresh your token. %s' % str(e)
        }
    else:
        @request.after
        def remember(response):
            # create the identity with the userid and updated user info
            identity = context.get_by_userid(userid).identity
            # create the updated token and set it in the response header
            request.app.remember_identity(response, request, identity)

        return {
            'status': 'success',
            'message': 'Token successfully refreshed'
        }


@App.json(model=UserCollection, name='logout')
def logout(context, request):
    """Log out the user."""

    @request.after
    def forget(response):
        request.app.forget_identity(response, request)

    return {
        'status': 'success'
    }


@App.json(model=UserModel, request_method='PATCH')
def update(context, request):
    error = None
    if 'password' in request.json.keys():
        error = {
            'status': 'error',
            'message': 'Changing password is not allowed through this method'
        }
    elif 'username' in request.json.keys():
        error = {
            'status': 'error',
            'message': 'Changing username is not allowed'
        }

    if error:
        @request.after
        def adjust_status(response):
            response.status = 422

        return error

    context.update(request.json)
    return {'status': 'success'}


@App.json(model=UserModel, name='roles')
def roles(context, request):
    return context.group_roles()


@App.json(model=UserModel, name='change_password',
          permission=permission.ChangePassword,
          request_method='POST')
def change_password(context, request):
    data = request.json
    error = None
    current_password = data.get('password', '')
    if not has_role(request, 'administrator'):
        if not context.validate(current_password, check_state=False):
            error = 'Invalid password'

    if not error and data['new_password'] != data['new_password_validate']:
        error = 'Password confirmation does not match'

    if error:
        @request.after
        def adjust_status(response):
            response.status = 422
        return {
            'status': 'error',
            'message': error
        }

    context.change_password(current_password, data['new_password'])
    return {'status': 'success'}



from morepath.publish import resolve_model as _resolve_model
from ..interfaces import ISchema
import jsl
import jsonobject
import dataclasses
from copy import copy
import typing
from datetime import datetime, date


def resolve_model(request):
    newreq = request.app.request_class(
        request.environ.copy(), request.app, path_info=request.path)
    context = _resolve_model(newreq)
    context.request = request
    return context


def jsonobject_property_to_jsl_field(
        prop: jsonobject.JsonProperty, nullable=False) -> jsl.BaseField:
    if isinstance(prop, jsonobject.DateProperty):
        return jsl.DateTimeField(name=prop.name, required=prop.required)
    if isinstance(prop, jsonobject.DateTimeProperty):
        return jsl.DateTimeField(name=prop.name, required=prop.required)
    if isinstance(prop, jsonobject.StringProperty):
        return jsl.StringField(name=prop.name, required=prop.required)
    if isinstance(prop, jsonobject.IntegerProperty):
        return jsl.IntField(name=prop.name, required=prop.required)
    if isinstance(prop, jsonobject.FloatProperty):
        return jsl.NumberField(name=prop.name, required=prop.required)
    if isinstance(prop, jsonobject.BooleanProperty):
        return jsl.BooleanField(name=prop.name, required=prop.required)
    if isinstance(prop, jsonobject.DictProperty):
        if prop.item_wrapper:
            subtype = jsonobject_to_jsl(
                prop.item_wrapper.item_type, nullable=nullable)
            return jsl.DocumentField(name=prop.name,
                                     document_cls=subtype,
                                     required=prop.required)
        return jsl.DictField(name=prop.name, required=prop.required)
    if isinstance(prop, jsonobject.ListProperty):
        if prop.item_wrapper:
            if isinstance(prop.item_wrapper, jsonobject.ObjectProperty):
                if issubclass(prop.item_wrapper.item_type, jsonobject.JsonObject):
                    subtype = jsl.DocumentField(
                        document_cls=jsonobject_to_jsl(prop.item_wrapper.item_type), nullable=nullable)
                elif isinstance(prop.item_wrapper.item_type, jsonobject.JsonProperty):
                    subtype = jsonobject_property_to_jsl_field(
                        prop.item_wrapper.item_type)
                else:
                    raise KeyError(prop.item_wrapper.item_type)
            elif isinstance(prop.item_wrapper, jsonobject.StringProperty):
                subtype = jsl.StringField(name=prop.name)
            elif isinstance(prop.item_wrapper, jsonobject.IntegerProperty):
                subtype = jsl.IntField(name=prop.name)
            elif isinstance(prop.item_wrapper, jsonobject.FloatProperty):
                subtype = jsl.NumberField(name=prop.name)
            elif isinstance(prop.item_wrapper, jsonobject.DictProperty):
                subtype = jsl.DictField(name=prop.name)
            else:
                raise KeyError(prop.item_wrapper)
            return jsl.ArrayField(items=subtype, required=prop.required)
        return jsl.ArrayField(name=prop.name, required=prop.required)

    raise KeyError(prop)


_marker = object()


def dataclass_get_type(field):
    metadata = {
        'required': _marker,
        'exclude_if_empty': False,
        'validators': []
    }
    metadata.update(field.metadata.get('morpfw', {}))

    origin = getattr(field.type, '__origin__', None)
    required = True
    if origin == typing.Union:
        if len(field.type.__args__) == 2:
            if field.type.__args__[1] == type(None):
                required = False
            typ = field.type.__args__[0]
    else:
        typ = field.type

    if metadata['required'] is _marker:
        metadata['required'] = required

    required = metadata['required']

    origin = getattr(typ, '__origin__', None)

    if origin == list:
        if getattr(typ, '__args__', None):
            return {
                'name': field.name,
                'type': list,
                'schema': field.type.__args__[0],
                'required': required,
                'metadata': metadata
            }
        else:
            return {
                'name': field.name,
                'type': list,
                'required': required,
                'metadata': metadata
            }

    return {
        'type': typ,
        'required': required,
        'metadata': metadata
    }


def dataclass_check_type(field, basetype):

    t = dataclass_get_type(field)

    if t['type'] == basetype:
        return t

    # wtf bool is a subclass of integer?
    if t['type'] == bool and basetype == int:
        return None

    if issubclass(t['type'], basetype):
        return t

    return None


def dataclass_field_to_jsl_field(
        prop: dataclasses.Field, nullable=False) -> jsl.BaseField:
    t = dataclass_check_type(prop, date)
    if t:
        return jsl.DateTimeField(name=prop.name, required=t['required'])
    t = dataclass_check_type(prop, datetime)
    if t:
        return jsl.DateTimeField(name=prop.name, required=t['required'])
    t = dataclass_check_type(prop, str)
    if t:
        return jsl.StringField(name=prop.name, required=t['required'])
    t = dataclass_check_type(prop, int)
    if t:
        return jsl.IntField(name=prop.name, required=t['required'])
    t = dataclass_check_type(prop, float)
    if t:
        return jsl.NumberField(name=prop.name, required=t['required'])
    t = dataclass_check_type(prop, bool)
    if t:
        return jsl.BooleanField(name=prop.name, required=t['required'])
    t = dataclass_check_type(prop, dict)
    if t:
        return jsl.DictField(name=prop.name, required=t['required'])
    t = dataclass_check_type(prop, ISchema)
    if t:
        subtype = jsonobject_to_jsl(
            t['schema'], nullable=nullable)
        return jsl.DocumentField(name=prop.name,
                                 document_cls=subtype,
                                 required=t['required'])

    t = dataclass_check_type(prop, list)
    if t:
        return jsl.ArrayField(name=prop.name, required=t['required'])

    t = dataclass_check_type(prop, typing.List)
    if t:
        if 'schema' not in t.keys():
            return jsl.ArrayField(name=prop.name, required=t['required'])

        if issubclass(t['schema'], ISchema):
            subtype = jsl.DocumentField(
                document_cls=jsonobject_to_jsl(
                    t['schema'], nullable=nullable))
        elif t['schema'] == str:
            subtype = jsl.StringField(name=prop.name)
        elif t['schema'] == int:
            subtype = jsl.IntField(name=prop.name)
        elif t['schema'] == float:
            subtype = jsl.NumberField(name=prop.name)
        elif t['schema'] == dict:
            subtype = jsl.DictField(name=prop.name)
        else:
            raise KeyError(t['schema'])
        return jsl.ArrayField(items=subtype, required=t['required'])

    raise KeyError(prop)


def _set_nullable(prop):
    if not prop.required:
        return jsl.OneOfField([prop, jsl.NullField(name=prop.name)])
    return prop


def dataclass_to_jsl(schema, nullable=False):
    attrs = {}

    class Options(object):
        additional_properties = True

    for attr, prop in schema.__dataclass_fields__.items():
        prop = dataclass_field_to_jsl_field(prop, nullable=False)
        if nullable:
            attrs[attr] = _set_nullable(prop)
        else:
            if not prop.required:
                attrs[attr] = prop
            else:
                if isinstance(prop, jsl.StringField):
                    if not prop.pattern:
                        prop.pattern = '.+'
                attrs[attr] = prop

    attrs['Options'] = Options
    Schema = type("Schema", (jsl.Document, ), attrs)

    return Schema


def jsonobject_to_jsl(schema, nullable=False):
    # output jsl schema from jsonobject schema
    attrs = {}

    class Options(object):
        additional_properties = True

    for attr, prop in schema._properties_by_attr.items():
        prop = jsonobject_property_to_jsl_field(prop, nullable=nullable)
        if nullable:
            attrs[attr] = _set_nullable(prop)
        else:
            if not prop.required:
                attrs[attr] = prop
            else:
                if isinstance(prop, jsl.StringField):
                    if not prop.pattern:
                        prop.pattern = '.+'
                attrs[attr] = prop

    attrs['Options'] = Options
    Schema = type("Schema", (jsl.Document, ), attrs)

    return Schema


def jsl_nullable(schema):
    return jsonobject_to_jsl(jsl_to_jsonobject(schema), nullable=True)


def jsl_field_to_jsonobject_property(
        prop: jsl.BaseField) -> jsonobject.JsonProperty:
    if isinstance(prop, jsl.DateTimeField):
        return jsonobject.DateTimeProperty(name=prop.name,
                                           required=prop.required)
    if isinstance(prop, jsl.StringField):
        return jsonobject.StringProperty(name=prop.name,
                                         required=prop.required)
    if isinstance(prop, jsl.IntField):
        return jsonobject.IntegerProperty(name=prop.name,
                                          required=prop.required)
    if isinstance(prop, jsl.DictField):
        return jsonobject.DictProperty(name=prop.name, required=prop.required)
    if isinstance(prop, jsl.NumberField):
        return jsonobject.FloatProperty(name=prop.name, required=prop.required)
    if isinstance(prop, jsl.BooleanField):
        return jsonobject.BooleanProperty(name=prop.name,
                                          required=prop.required)
    if isinstance(prop, jsl.DocumentField):
        if prop.document_cls:
            subtype = jsl_to_jsonobject(prop.document_cls)
            return jsonobject.DictProperty(name=prop.name,
                                           item_type=subtype,
                                           required=prop.required)
        return jsonobject.DictProperty(name=prop.name, required=prop.required)
    if isinstance(prop, jsl.ArrayField):
        if prop.items:
            if isinstance(prop.items, jsl.DocumentField):
                subtype = jsl_to_jsonobject(prop.items.document_cls)
            elif isinstance(prop.items, jsl.BaseField):
                subtype = jsl_field_to_jsonobject_property(prop.items)
            else:
                raise KeyError(prop.items)
            return jsonobject.ListProperty(item_type=subtype,
                                           required=prop.required)
        return jsonobject.ListProperty(name=prop.name, required=prop.required)

    raise KeyError(prop)


def jsl_to_jsonobject(schema):
    # output jsonobject schema from jsl schema
    attrs = {}
    for attr, prop in schema._fields.items():
        prop.name = attr
        attrs[attr] = jsl_field_to_jsonobject_property(prop)

    Schema = type("Schema", (jsonobject.JsonObject, ), attrs)

    return Schema


def generate_default(schema):
    data = {}
    if isinstance(schema, jsl.DocumentField):
        schema = schema.document_cls
    for n, f in schema._fields.items():
        if isinstance(f, jsl.DocumentField):
            data[n] = generate_default(f)

        else:
            data[n] = f.get_default()
            if data[n] is None:
                if isinstance(f, jsl.StringField):
                    data[n] = None
                elif (isinstance(f, jsl.IntField) or
                      isinstance(f, jsl.NumberField)):
                    data[n] = None
                elif isinstance(f, jsl.DictField):
                    data[n] = {}
                elif isinstance(f, jsl.ArrayField):
                    data[n] = []

    return data



from jsonschema.validators import Draft4Validator
from jsonschema import validate, ValidationError
from .util import jsl_nullable
from .util import dataclass_to_jsl
from .const import SEPARATOR
from . import permission
from . import signals
from .log import logger
from rulez import validate_condition
from morepath import reify
from DateTime import DateTime
from uuid import uuid4
from transitions import Machine
import copy
from .errors import StateUpdateProhibitedError, AlreadyExistsError, BlobStorageNotImplementedError
from .errors import UnprocessableError
from ..interfaces import IModel, ICollection
import json


ALLOWED_SEARCH_OPERATORS = [
    'and', 'or', '==', 'in',
    '~', '!=', '>', '<', '>=',
    '<='
]

_marker = object()


class Collection(ICollection):

    create_view_enabled = True
    search_view_enabled = True
    aggregate_view_enabled = True

    exist_exc = AlreadyExistsError

    @property
    def schema(self):
        raise NotImplementedError

    def __setitem__(self, key, value):
        self.data[key] = value

    def __getitem__(self, key):
        return self.data[key]

    def __delitem__(self, key):
        del self.data[key]

    def __init__(self, request, storage, data=None):
        self.request = request
        self.app = request.app
        self.storage = storage
        self.data = None
        if data:
            self.data = request.app.get_dataprovider(self.schema, data,
                                                     self.storage)

    def search(self, query=None, offset=0, limit=None, order_by=None,
               secure=False):
        objs = self._search(query, offset, limit, order_by, secure)
        if secure and limit:
            nextpage = {'query': query, 'offset': offset +
                        limit, 'limit': limit, 'order_by': order_by}
            while len(objs) < limit:
                nextobjs = self._search(secure=True, **nextpage)
                if len(nextobjs) == 0:
                    return list(objs[:limit])
                nextpage['offset'] = nextpage['offset'] + limit
                objs = objs + nextobjs
        return objs

    def _search(self, query=None, offset=0, limit=None, order_by=None,
                secure=False):
        if query:
            validate_condition(query, ALLOWED_SEARCH_OPERATORS)
        objs = self.storage.search(
            query, offset=offset, limit=limit, order_by=order_by)

        if secure:
            objs = list([obj for obj in objs if self.request.app.permits(
                self.request, obj, permission.View)])
        return list(objs)

    def aggregate(self, query=None, group=None, order_by=None):
        if query:
            validate_condition(query, ALLOWED_SEARCH_OPERATORS)
        objs = self.storage.aggregate(query, group=group, order_by=order_by)
        return list(objs)

    def create(self, data):
        identifier = self.app.get_default_identifier(
            self.schema, data, self.request)
        if identifier and self.get(identifier):
            raise self.exist_exc(identifier)
        obj = self._create(data)
        obj.set_initial_state()
        self.request.app.signal_publish(self.request,
                                        obj, signals.OBJECT_CREATED)
        return obj

    def _create(self, data):
        data = self.storage.set_schema_defaults(data)
        return self.storage.create(data)

    def get(self, identifier):
        if isinstance(identifier, list) or isinstance(identifier, tuple):
            identifier = self.request.app.join_identifier(*identifier)
        return self.storage.get(identifier)

    def get_by_uuid(self, uuid):
        return self.storage.get_by_uuid(uuid)

    def json(self):
        return {
            'schema': dataclass_to_jsl(self.schema).get_schema(ordered=True),
            'links': self.links()
        }

    def links(self):
        request = self.request
        links = [{'rel': 'create',
                  'href': request.link(self),
                  'method': 'POST'},
                 {'rel': 'search',
                  'href': request.link(self, '+search')}]
        links += self._links()
        return links

    def _links(self):
        return []


class Model(IModel):

    linkable = True
    update_view_enabled = True
    delete_view_enabled = True

    blobstorage_field = 'blobs'
    blob_fields = []
    protected_fields = [
        'id',  'blobs', 'state',  'xattrs',
        'modified', 'created', 'uuid',
        'creator', 'deleted'
    ]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __getitem__(self, key):
        return self.data[key]

    def __delitem__(self, key):
        del self.data[key]

    def __dict__(self):
        return self.data.as_dict()

    @property
    def statemachine_view_enabled(self):
        if self.state_machine():
            return True
        return False

    @property
    def xattr_view_enabled(self):
        if self.xattrprovider():
            return True
        return False

    @property
    def schema(self):
        raise NotImplementedError

    @property
    def identifier(self):
        if self._cached_identifier:
            return self._cached_identifier
        res = []
        for f in self.app.get_identifierfields(self.schema):
            d = self.data.get(f)
            if d is not None:
                d = str(d)
            res.append(d)
        if None in res:
            identifier = self.app.get_default_identifier(
                self.schema, self.data, self.request)
            if identifier is None:
                return None
            if isinstance(identifier, list) or isinstance(identifier, tuple):
                identifier = self.app.join_identifier(*identifier)
            self.storage.set_identifier(self.data, identifier)
            self._cached_identifier = identifier
            return identifier
        separator = self.request.app.get_compositekey_separator()
        identifier = separator.join(res)
        self._cached_identifier = identifier
        return identifier

    @property
    def uuid(self):
        uuid_field = self.app.get_uuidfield(self.schema)
        return self.data[uuid_field]

    def __init__(self, request, storage, data):
        self.request = request
        self.storage = storage
        self.app = request.app
        self.data = request.app.get_dataprovider(self.schema, data,
                                                 self.storage)
        self._cached_identifier = None
        super().__init__(request, storage, data)

    def update(self, newdata, secure=False):

        if secure:
            if 'state' in newdata:
                raise StateUpdateProhibitedError()
            for protected in self.protected_fields:
                if protected in newdata.keys():
                    raise UnprocessableError(
                        "%s is not allowed to be updated in this context" % protected)

        data = self._raw_json()
        data.update(newdata)
        schema = dataclass_to_jsl(
            self.schema, nullable=True).get_schema(ordered=True)
        validate(data, schema)
        self.storage.update(self.identifier, data)
        self.request.app.signal_publish(
            self.request, self, signals.OBJECT_UPDATED)

    def delete(self):
        self.request.app.signal_publish(
            self.request, self, signals.OBJECT_TOBEDELETED)
        self.storage.delete(self.identifier, model=self)

    def save(self):
        if self.data.changed:
            data = self._raw_json()
            schema = dataclass_to_jsl(
                self.schema, nullable=True).get_schema(ordered=True)
            validate(data, schema)
            self.storage.update(self.identifier, data)

    def _raw_json(self):
        schema = dataclass_to_jsl(
            self.schema, nullable=True).get_schema(ordered=True)
        jsondata = self.app.get_jsonprovider(self.data)
        jsondata = self.rules_adapter().transform_json(copy.deepcopy(jsondata))
        try:
            validate(jsondata, schema)
        except ValidationError as e:
            logger.warn('%s(%s) : %s' % (self.schema.__name__,
                                         '/'.join(list(e.path)), e.message))
        return jsondata

    def _json(self):
        return self._raw_json()

    def json(self):
        if self.linkable:
            return {
                'data': self._json(),
                'links': self.links()
            }
        return self._json()

    def as_json(self):
        return self.data.as_json()

    def as_dict(self):
        return self.data.as_dict()

    def links(self):
        links = []
        links.append({
            'rel': 'self',
            'href': self.request.link(self)
        })
        if self.update_view_enabled:
            links.append({
                'rel': 'update',
                'href': self.request.link(self),
                'method': 'PATCH'
            })
        if self.delete_view_enabled:
            links.append({
                'rel': 'delete',
                'href': self.request.link(self),
                'method': 'DELETE'
            })
        if self.statemachine_view_enabled:
            links.append({
                'rel': 'statemachine',
                'href': self.request.link(self, '+statemachine'),
                'method': 'POST'
            })
        links += self._links()
        links += self.rules_adapter().links()
        return links

    def _links(self):
        return []

    def rules_adapter(self):
        return self.app.get_rulesadapter(self)

    def state_machine(self):
        return self.statemachine()

    def statemachine(self):
        if self.app.get_statemachine.by_args(self).all_matches:
            return self.app.get_statemachine(self)
        return None

    def xattrprovider(self):
        if self.app.get_xattrprovider.by_args(self).all_matches:
            return self.app.get_xattrprovider(self)
        return None

    def set_initial_state(self):
        self.state_machine()

    def _blob_guard(self, field):
        if self.blobstorage_field not in self.schema.__dataclass_fields__.keys():
            raise BlobStorageNotImplementedError(
                'Object does not implement blobs store')
        if field not in self.blob_fields:
            raise BlobStorageNotImplementedError(
                'Field %s not allowed for blobstorage' % field)

    def put_blob(self, field, fileobj, filename, mimetype=None, size=None, encoding=None):
        self._blob_guard(field)
        blob_data = self.data[self.blobstorage_field] or {}
        existing = blob_data.get(field, None)
        blob = self.storage.put_blob(
            fileobj, filename, mimetype, size, encoding)
        blob_data[field] = blob.uuid
        self.update({self.blobstorage_field: blob_data})
        if existing:
            self.storage.delete_blob(existing)
        return blob

    def get_blob(self, field):
        if not self.data[self.blobstorage_field]:
            return None
        uuid = self.data[self.blobstorage_field][field]
        blob = self.storage.get_blob(uuid)
        return blob

    def delete_blob(self, field):
        uuid = self.data[self.blobstorage_field][field]
        self.storage.delete_blob(uuid)


class StateMachine(object):

    @property
    def states(self):
        raise NotImplementedError

    @property
    def transitions(self):
        raise NotImplementedError

    def __init__(self, context):
        self._context = context
        self._request = context.request
        self._app = context.request.app
        initial = self.state or self.states[0]
        self._machine = Machine(
            model=self, transitions=self.transitions,
            states=self.states, initial=initial)

    @property
    def state(self):
        try:
            return self._context.data['state']
        except KeyError:
            return None

    @state.setter
    def state(self, val):
        self._context.data['state'] = val

    def get_triggers(self):
        return [i for i in self._machine.get_triggers(self.state) if not i.startswith('to_')]


class XattrProvider(object):

    @property
    def schema(self):
        raise NotImplementedError

    def __init__(self, context):
        self.context = context
        self.request = context.request
        self.app = context.request.app

    def jsonschema(self):
        schema = dataclass_to_jsl(
            self.schema, nullable=True).get_schema(ordered=True)
        return {
            'schema': schema
        }

    def as_json(self):
        raise NotImplementedError

    def as_dict(self):
        raise NotImplementedError

    def process_update(self, newdata: dict):
        data = self.as_json()
        data.update(newdata)
        schema = dataclass_to_jsl(
            self.schema, nullable=True).get_schema(ordered=True)
        validate(data, schema)
        self.update(data)

    def update(self, newdata: dict):
        raise NotImplementedError

    def __setitem__(self, key, value):
        raise NotImplementedError

    def __getitem__(self, key):
        raise NotImplementedError

    def __delitem__(self, key):
        raise NotImplementedError

    def get(self, key, default=_marker):
        raise NotImplementedError



import morpfw


class App(morpfw.SQLApp):
    pass


class MountedApp(morpfw.SQLApp):
    pass


@App.mount(app=MountedApp, path='/api/v1')
def mount_app(app):
    return MountedApp()


if __name__ == '__main__':
    morpfw.run(App, {})
